;-- EaDIO8Test (Tests fuer E/A-Modul "EA7 DIO8 V1.0" mit ARM7 & CANopen) ------
; ( auch fr "EA7-COMPACT MONO16" geeignet,
;        seit 2016 auch fr den "IO16-BCCO" = "Buskoppler mit CANopen",
;        seit 2024 auch fr das "EA550-8IO" mit 8 digitalen Ausgngen
;                   und -alternativ- 8 analogen oder digitalen Eingngen ).
; Dieses Script fr WB's CAN-Tester zeigt zunchst Informationen ber das
; angeschlossene CANopen-Gert an, und versucht dann -falls geeignete
; Test-Prozeduren fr Digital- und Analog-I/O existieren- diese zu starten.
;
; Programm starten mit F9 bzw. "Run"..."Start"  !
;
; Autor: Wolfgang Buescher     Stand: 2024-03-13
;   Ablage des Originals: //WoBu5/c/CBproj/CanTest/Programs/EaDIO8Test.cts .
;   Mit Test der Funktionalitaet der Firmware nach DS301 V4.0 (!).
; Wegen diverser Sonderfunktionen NUR MIT CAN-TESTER AB V1.73 lauffhig !
; Falls sich das Programm merkwrdig benimmt, bitte den CAN-Tester updaten .
; Im Hauptmen des CAN-Testers unter "Tools".."SDO Client" diesen AKTIVIEREN,
;    wenn dort nicht bereits "SDO Client (active)" steht .
;
; Einstellung der PORT-RICHTUNGEN fr "EA7-DIO8" am ENDE der Datei
;  oder mit den Buttons "Ausgnge aktivieren" / .. "deaktivieren"  .
;

Neustart::
can.reset        // CAN-Controller ruecksetzen (hilft gegen BUS-OFF und Bug im Peak-Treiber)
@NumErrors := 0  // clear error counter
@tx_interval=0.001  // set TX interval, also limits execution speed: SECONDS per line
OnError goto Verflixt // define error handler (jump if "something" goes wrong)


;-- Einstellung Variablen ---------------------------------------------------
OpenWindow( "E/A-DIO8-Test Vorbereitung",w=320,h=100)
 SetFont("Arial",8,1)
 print N,J,P,S,T,Y,L  ; Test: press F2 with cursor in this line to check settings !
 print "CAN: %ld kBit/sec",can.baudrate
NodeId := 1;  // Node-ID (==DIP-Schalter-Adresse, beim Terminal im Systemmenue)
J  := 384+NodeId;   // Identifier for PDO I/O ->Master (Inputs,Lenze: 767+N, DS401:384+N)
P1 := 0x200+NodeId; // Identifier for TPDO1 Master -> I/O-module (DS401:0x200+N)
P2 := 0x300+NodeId; // Identifier for TPDO2 Master -> I/O-module (DS401:0x300+N)
P3 := 0x400+NodeId; // Identifier for TPDO3 Master -> EA (Analog-Out, or >128 digital Outputs)

@S=1536+NodeId    ; SDO(rx) des E/A-Master-Moduls
@T=1408+NodeId    ; SDO(tx) des E/A-Master-Moduls (!)
@Y= 1             ; Option: 0=ohne SYNC  1=mit SYNC
@L= 1             ; Option: 0=Laufl.aus  1=Laufl.an
@sdo.node(NodeId) ; Node-ID fuer SDO-Zugriffe mit sdo.read() und write()
@print "NODE-ID fuer SDO-Zugriffe = %ld", sdo.node

; @set_scope.timing 20
; @set_scope(5).range 19800 20200
; @set_scope(5).range -32768 32767

 000 81 00    ; NMT-command: any state -> Initialisation (Reset),
              ;  must be "acknowledged" with a "bootup message" .
 delay(1.0)   ; Caution, the SYSTEC CANopen stack needs some time to think now !

CloseWindow()

;@goto LaufLicht

;----- 'Data Types' and 'Defstructs' ----------------------
@goto SkipDefStructs
  print "DefStruct_PDOCommPar=$%08lX",sdo.r($0020.0) ; "DEFSTRUCT", kind of type-definition readable via SDO ...
  print "typeof(PDOCommPar.CANid)=%ld",   sdo.r($0020.1) ; should be '7' = UINT32
  print "typeof(PDOCommPar.TranType)=%ld",sdo.r($0020.2) ; should be '5' = UINT8
  print "typeof(PDOCommPar.InhTime)=%ld", sdo.r($0020.3) ; should be '6' = UINT16
  print "typeof(PDOCommPar.Reserved)=%ld",sdo.r($0020.4) ; should be '5' = UINT8
  print "typeof(PDOCommPar.EventTmr)=%ld",sdo.r($0020.5) ; should be '6' = UINT16
  print "typeof(PDOCommPar.SyncStrt)=%ld",sdo.r($0020.6) ; should be '5' = UINT8
SkipDefStructs:


;----- Communication Profile (DS301 V4) -------------------
OpenWindow( "E/A-DIO8-Test Schritt 1 : Identifikation", w=600, h=500 )
 ClearScreen() : SetFont("Arial",20,1) : print("wichtige CANopen-Objekte / Konfiguration")
 SetFont("Arial",8,1) : MoveTo(20,40)
 print "Standard-Kommunikations-Profil" : SetFont("Arial",8,0)
 print "DevType=$%08lX",sdo.r($1000.0)
; print "NonExistingTest=$%08lX",sdo.r($1000.1)      ; dies muss mit einem "Abort Code" bestraft werden (denn 0x1000.01 existiert nicht)
 print "ErrorReg=$%08lX",sdo.r($1001.0)
 ;@print "ManufStatusReg=$%08lX",sdo.r($1002.0)     ; nicht implementiert
 @N_Errors = sdo.r($1003.0)  ; Number of errors (that have been signalled via EMERGENCY) ?
 print "Number of errors=$%08lX",N_Errors
 if( N_Errors > 0 )
    print "Newest error=$%08lX",sdo.r($1003.1)      ; Most recent error (signalled via EMERGENCY)
 endif 
 ;@print "NumberOftotalPDOs=$%08lX",sdo.r($1004.0)  ; Objekt 0x1004 hat man in DS301 V4 ...
 ;@print "NumberOfsynchPDOs=$%08lX",sdo.r($1004.1)  ; ... schlicht und einfach RAUSGESCHMISSEN !
 ;@print "NumberOfasyncPDOs=$%08lX",sdo.r($1004.2)
 print "CobID_Sync=$%08lX",sdo.r($1005.0)
 print "CommCyclePer=%ld",sdo.r($1006.0)            ; "Communication Cycle" aka "SYNC Transmit Cycle"
 print "GuardTime=$%08lX",sdo.r($100C.0)
 print "LifeTimeFactor   =$%08lX",sdo.r($100D.0)
 ;@print "CobIdTimestamp =$%08lX",sdo.r($1012.0)    ; not projected
 print "CobIdEmergency   =$%08lX",sdo.r($1014.0)
 print "EmgcyInhibitTime =$%08lX",sdo.r($1015.0)
 print "ProducerHeartbeat=$%08lX",sdo.r($1017.0)   ; Note: this is an ARRAY, subindex 0 = "number of elements" !
 print "IdentityObjectSubis=%08lX",sdo.r($1018.0)  ; Note: this is a RECORD, subindex 0 = "number of members" !
 print "RPDO1_CommPar=$%08lX",sdo.r($1400.0) ; highest subindex
 print "RPDO1_CobID  =$%08lX",sdo.r($1400.1)
 print "RPDO1_TranType=$%08lX",sdo.r($1400.2)
 print "RPDO1_InhiTime=$%08lX",sdo.r($1400.3)
;@print "RPDO1_CMS_prio=$%08lX",sdo.r($1400.4) ; Obj 0x1400.4 killed by CiA
 print "RPDO1_EvtTimer=$%08lX",sdo.r($1400.5)
 print "RPDO2_CommPar=$%08lX",sdo.r($1401.0)
 print "RPDO2_CobID  =$%08lX",sdo.r($1401.1)
 print "RPDO2_TranType=$%08lX",sdo.r($1401.2)
 print "RPDO2_InhiTime=$%08lX",sdo.r($1401.3)
;@print "RPDO2_CMS_prio=$%08lX",sdo.r($1401.4) ; Obj 0x1401.4 killed by CiA
 print "RPDO2_EvtTimer=$%08lX",sdo.r($1401.5)
 print "RPDO3_CobID = $%08lX",sdo.r($1402.1)
 print "TPDO1_CobID = $%08lX",sdo.r($1800.1)
 print "TPDO2_CobID = $%08lX",sdo.r($1801.1)
 print "TPDO3_CobID = $%08lX",sdo.r($1802.1)
 print "TPDO4_CobID = $%08lX",sdo.r($1803.1,U32,0)
 print "TPDO5_CobID = $%08lX",sdo.r($1804.1,U32,0)
 print "TPDO6_CobID = $%08lX",sdo.r($1805.1,U32,0)
 print "TPDO7_CobID = $%08lX",sdo.r($1806.1,U32,0)
 print "TPDO8_CobID = $%08lX",sdo.r($1807.1,U32,0)



 // Ab hier darf der Bediener "abkrzen" wenn er will, 
 // darum werden jetzt die Buttons eingeblendet :
 SetButton( 1, "&Beenden",   goto Beenden   )
 SetButton( 2, "&Ausgnge aktivieren",   goto OutputsActive  )
 SetButton( 3, ".. &deaktivieren", goto OutputsPassive )
 SetButton( 4, "&Optionen", goto OtherOptions )
 SetButton( 5, "&Weiter >>", goto DigIoTest )

 SetFont("Arial",8,1) : MoveTo(210,40)
 print "Modul-Identifikation a la CANopen" : SetFont("Arial",8,0)
 print "Device name = %s", sdo.r($1008.0) ; device name (as "visible string")
 print "Hardware version = %s", sdo.r($1009.0) ; hardware version ("visible string")
 print "Software version = %s", sdo.r($100A.0) ; software version ("visible string")
 print "Serial Number    = %ld",sdo.r($1018.4) ; serial number (part of "identity object")
; print "Must Abort Here  : %ld",sdo.r($1018.5) ; non-existing member -> abort code ?


 SetFont("Arial",8,1)
 print "\nDigital input setup (DS401)" : SetFont("Arial",8,0)
 @NDI := sdo.r($6000.0,U8,0) // read number of digital Inputs (0 if access fails)
 print "NumDigInputGroups_8Bit = %ld", NDI
 if (NDI>=1)            // if digital inputs 1..8 exist :
   print "DigitalInputsA = %08b",sdo.r($6000.1)   // show state of inputs (first 8-bit-group)
   print "DinInterruptMskA = %08b",sdo.r($6006.1) // show trigger mask
 endif
 if (NDI>=2)            // if digital inputs 9..16 exist :
   print "DigitalInputsB = %08b",sdo.r($6000.2)   // show state of inputs
   print "DinInterruptMskB = %08b",sdo.r($6006.2) // show trigger mask
 endif
 if (NDI>=3)            // if digital inputs 17..24 exist :
   print "DigitalInputsC = %08b",sdo.r($6000.3)   // show state of inputs
   print "DinInterruptMskC = %08b",sdo.r($6006.3) // show trigger mask
 endif
 if (NDI>=4)            // if digital inputs 25..32 exist :
   print "DigitalInputsD = %08b",sdo.r($6000.4)   // show state of inputs
   print "DinInterruptMskD = %08b",sdo.r($6006.4) // show trigger mask
 endif
 if (NDI>=5)            // if digital inputs 33..40 exist :
   print "DigitalInputsE = %08b",sdo.r($6000.5)   // show state of inputs
   print "DinInterruptMskE = %08b",sdo.r($6006.5) // show trigger mask
 endif

 print "DinInterruptEnable= %ld",sdo.r($6005.0) ; read "global interrupt-enable"-flag for d.inputs

 SetFont("Arial",8,1)
 print "\nDigital output setup (DS401)" : SetFont("Arial",8,0)
 @NDO := sdo.r($6200.0,U8,0) // read number of digital OUTPUTS (0 if access fails)
 print "NumDigOutputGroups_8Bit =%ld",NDO        // show number of digital output groups
 if (NDO>=1)            // if digital outputs 1..8 exist :
   print "DigitalOutputsA = %08b",sdo.r($6200.1) // show state of digital outputs
   print "DigOutputMaskA  = %08b",sdo.r($6208.1,U8,0) // Filter Mask Digital Output 8-bit
   // (were ok after POWER-ON, but got lost during NODE-RESET somehow in EA7-MONO16) !
   print "PortDirectionsA = %08b",sdo.r($5FF5.1) // show their port-directions
   // (Die PORT-RICHTUNGEN knnen per SDO am Ende dieser Datei gendert werden)
 endif
 if (NDO>=2)            // if digital outputs 9..16 exist :
   print "DigitalOutputsB = %08b",sdo.r($6200.2) // show state of digital outputs
   print "DigOutputMaskB  = %08b",sdo.r($6208.2,U8,0) // Filter Mask Digital Output 8-bit
   print "PortDirectionsB = %08b",sdo.r($5FF5.2) // show their port-directions
 endif

 SetFont("Arial",8,1) : MoveTo(410,40)
 print "Analog input setup (DS401)" : SetFont("Arial",8,0)
 @NAI := sdo.r($6401.0,U8,0) // read number of ANALOG inputs (0 if access fails)
 print "NumAnalogInputs =%ld",NAI     // show number of analog Inputs
 if (NAI>=1)            // only if 1st analog input exists...
   print "AnalogInput1 =%ld",sdo.r($6401.1) // show 1st 16-bit-analog Input
 endif
 if (NAI>=2)            // only if 2nd analog input exists...
   print "AnalogInput2 =%ld",sdo.r($6401.2) // show 2nd 16-bit-analog Input
 endif
 if (NAI>=3)            // only if 3rd analog input exists...
   print "AnalogInput3 =%ld",sdo.r($6401.3) // show 3rd 16-bit-analog Input
 endif
 if (NAI>=4)             // only if 4th analog input exists...
   print "AnalogInput4 =%ld",sdo.r($6401.4) // show 4th 16-bit-analog Input
 endif
 if (NAI>0)
   print "AinTrigEnable= %ld", sdo.r($6423.0) //"global interrupt-enable"-flag for AIs
 endif
 N := sdo.r($6422.0,U8,0) // read the number of analog input trigger "sources"
 print "Analog Trigger Sources = %ld",N
 if (N>=1) print "AinIntrSrc1 = 0x%08lX",sdo.r($6422.1)
 if (N>=2) print "AinIntrSrc2 = 0x%08lX",sdo.r($6422.2)
 if (N>=3) print "AinIntrSrc3 = 0x%08lX",sdo.r($6422.3)

 @N:=sdo.r($6421.0,U8,0) // read the number of analog input trigger "selections"
 print "Analog Trigger Selections = %ld",N
 if (N>=1)           // only if analog trigger supported for this channel:
   print "AinTrigSelN = %ld", sdo.r($6421.0) ; number of analog input trigger selections
   print "AinTrigSel1 = %08b",sdo.r($6421.1) ; analog input trigger selection #1
   print "AinTrigLevLo1 = %ld",sdo.r($6425.1) ; analog input trigger level LO, ch #1
   print "AinTrigLevHi1 = %ld",sdo.r($6424.1) ; analog input trigger level HI, ch #1
   print "AinTrigDelta1 = %ld",sdo.r($6426.1) ; analog input trigger delta, ch #1
 endif

 SetFont("Arial",8,1)
 print "\nAnalog output setup (DS401)" : SetFont("Arial",8,0)
 @NAO := sdo.r($6411.0,U8,0) // read number of ANALOG OUTPUTS (0 if access fails)
 print "NumAnalogOutputs =%ld",NAO   // show number of analog Outputs
 if (NAO>=1)     // only if 1st analog output exists...
   print "AnalogOutput1 =%ld",sdo.r($6411.1) // show 1st analog Output
 endif
 if (NAO>=2)     // only if 2nd analog output exists...
   print "AnalogOutput2 =%ld",sdo.r($6411.2) // show 2nd analog Output
 endif

Warten: 
  goto Warten  // Diese nette Endlosschleife wird per Button-Event verlassen !
CloseWindow()


OutputsActive:: // Aktivieren der DIGITALEN OUTPUTS per CANopen
                // DIES IST NORMALERWEISE NICHT DER AUSLIEFERUNGSZUSTAND !
 OpenWindow( "EA-Modul wird umkonfiguriert...",w=320,h=150)
 SetFont("Arial",10,1)
 print "Digitale Ausgnge werden AKTIVIERT ..."
 SetFont("Arial",10,0)
 print " Bitte vor der Auslieferung, wenn gewnscht,"
 print " die digitalen Ausgnge wieder DEAKTIVIEREN"
 print "   (nachdem die Ausgnge getestet wurden) !"
 sdo.w($5FF5.1,U8,0xFF) ; ENABLE first 8 digital OUTPUTS
 sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)
 000 81 00    ; Node Reset
 delay(1.0)   ; Caution, the SYSTEC CANopen stack needs some time to think now !
 print "Port-Richtungen Gruppe 1 = %08b",sdo.r($5FF5.1)
 delay(2.0)
 goto Neustart  // Zur Kontrolle ("Port-Richtungs-Register" anzeigen)

 // Sonderwnsche ? Hier gibt's noch andere Digital-Ausgangs-Konfigurationen:
 sdo.w($5FF5.1,U8,0x0F) ; 00001111(binaer) : Vier Inputs (MSBs), vier Outputs (LSBs)
 sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)


OutputsPassive:: // Deaktivieren der DIGITALEN OUTPUTS per CANopen
                // Dies ist (meistens) der AUSLIEFERUNGSZUSTAND des Moduls !
 OpenWindow( "EA-Modul wird umkonfiguriert...",w=320,h=150)
 SetFont("Arial",10,0)
 print "Digitale Ausgnge werden deaktiviert ..."
 sdo.w($5FF5.1,U8,0x00) ; DISABLE first 8 digital OUTPUTS
 sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)
 000 81 00                        ; Node Reset
 delay(1.0)   ; Caution, the SYSTEC CANopen stack needs some time to think now !
 print "Port-Richtungen Gruppe 1 = %08b",sdo.r($5FF5.1)
 delay(2.0)
 goto Neustart  // Zur Kontrolle ("Port-Richtungs-Register" anzeigen)



OtherOptions:: // PDO-Mapping anzeigen, Default-Settings wiederherstellen, etc
 OpenWindow( "PDO-Liste und spezielle Optionen", w=640, h=720 )
 SetButton( 1, "<< &Neustart", goto Neustart  )
 SetButton( 2, "&Beenden",     goto Beenden   )
 SetButton( 3, "&Restore Defaults",   goto RestoreDefaults )
 SetButton( 4, "&Weiter >>",   goto DigIoTest )
 ClearScreen() : SetFont("Arial",20,1) : print("Receive-PDOs") : SetFont("Courier New",12,0);
 print( "CommPar Cob-ID   Mapped objects.." );
 CommPar := 0x1400; // OD-index for the first RPDO communication parameter
 repeat
   gosub PrintPdoMapping;
   CommPar := CommPar+1;
 until CommPar > 0x1407;
 print;
 SetFont("Arial",20,1) : print("Transmit-PDOs") :  SetFont("Courier New",12,0);
 print( "CommPar Cob-ID   Mapped objects.." );
 @CommPar = 0x1800; // OD-index for first TPDO communication parameter
 repeat             // (transmitted from device, received by CAN-tester)
   gosub PrintPdoMapping;
   CommPar = CommPar+1;
 until CommPar > 0x1807; 
  // Prepare the "live PDO display" :
  CommPar = 0x1800; // OD-index for first TPDO communication parameter
  repeat            // (transmitted from device, received by CAN-tester)
    PdoCobId[CommPar] := sdo.r( (CommPar).01, U32, 0); // store COP-IDs in a Pythonish array or list
    CommPar = CommPar+1;
  until CommPar > 0x1807;
  print;
  SetFont("Arial",20,1) : print("Data received from PDO") :  SetFont("Courier New",12,0);
  CX := cursor.x;    // save output cursor position for later
  CY := cursor.y;
  000 01 00  ; enter OPERATIONAL state, start PDOs. Only a TRANSITION sends all PDOs with tr-type 255

PdoDisplayLoop:   // Warteschleife mit Anzeige der empfangenen PDOs ...
  MoveTo(CX, CY);    // back to the screen position near 'Data received from PDO'
  CommPar := 0x1800; // OD-index for first TPDO communication parameter
  repeat             // (transmitted from device, received by CAN-tester)
     CobId := PdoCobId[CommPar]; // ex:  sdo.r( (CommPar).01, U32, 0);
     // Wegen der schwachsinnigen Default-Einstellungen in DS401 V3.0 werden ANALOGE EINGNGE
     // per Default nur mit den TOTAL BESCHEUERTEN RTRs (Remote Transmit Requests) per PDO
     // gesendet. Ohne UMKONFIGURATION (per SDO) laufen die "Analog-PDOs" *NICHT* . 
     // Versuchen wir mal, das E/A-Modul per RTR(!) zum Senden von TPDO2,3,4 zu motivieren:
     if( (CommPar >= 0x1801) && (CommPar <= 0x1805) )
        (CobId) r XX XX XX XX XX XX XX XX  // send a dreadful REMOTE TRANSMIT REQUEST for 'analog PDO'
     endif;
     if( canrx.update_latch( CobId ) ) // try to capture the newest CAN message with this ID
        // Note: Individual message counters for each CAN-message-ID only work if
        // the CAN tester's display is running, and in mode "sorted" (by ID) !
        print("ID=%08lX: n=%04ld", CobId, canrx.latch.count; );
        ByteIdx := 0;
        PdoLen  := canrx.latch.len;  // length of the actually received PDO in bytes
        if( PdoLen > 0 )
           repeat
              Data :=  canrx.latch( 8*ByteIdx, U8 );
              print( " %02lX",Data; ); // show 8 hex-bytes (no NEW LINE yet)
              ByteIdx := ByteIdx + 1;
           until ByteIdx >= PdoLen;
           ClearEOL();     // clear old garbage (up to the end of the text line)
        else // received an EMPTY PDO, or no PDO at all yet ..
           print( " --"; );
           ClearEOL(red); 
        endif;
        print;  // only ONE new line per CAN message
    endif; // successfully 'latched' a CAN message from the RX-FIFO ?
    CommPar := CommPar+1; // try the next PDO 'communication parameter' in the next loop
  until CommPar > 0x1807;
  goto PdoDisplayLoop // wird nur per Button-Event verlassen !


PrintPdoMapping: // Lists all CANopen-objects currently mapped into a PDO .
  // [in] CommPar: e.g. 0x1400 = OD-index for first RPDO communication parameter, etc.
  @tx_interval=0.001  // set TX interval, also limits execution speed: SECONDS per line
  print("0x%04lX ",CommPar;);  // trailing semicolon in arg-list suppresses NEW LINE
  print(" %08lX",sdo.r( (CommPar).01, U32, 0) ; ); // "COB-ID" with invalid-flag in bit 31
  NumObjsInPDO := sdo.r( (CommPar+0x200).0, U32, 0);  // how many objects are 'mapped'  ?
  print(" [%d]",NumObjsInPDO ; );  // "number of objects mapped into a PDO"
  MappingSubidx := 1; // begin listing the "PDO mapping object" here
  repeat
     print(" %08lX",sdo.r( (CommPar+0x200).(MappingSubidx), U32, 0) ; );
     MappingSubidx = MappingSubidx+1;
  until MappingSubidx > NumObjsInPDO;
  print; // emit new line
return; 

GetSizeOfPDO: // added 2024-03 to find out the NUMBER OF BYTES to send in a PDO
  // [in] CommPar: e.g. 0x1400 = OD-index for first RPDO communication parameter, etc.
  // [out]  NumBytesMapped[CommPar] = Array with PDO-sizes. Note that the INDEX into
  //                      this array is the PDO's *communication paramter* (OD-index).
  NumObjsInPDO := sdo.r( (CommPar+0x200).0, U32, 0);  // how many objects are 'mapped'  ?
  MappingSubidx := 1; // begin summing up the size of mapped objects here..
  NumBitsMapped := 0; // LATER, the total number of BITS mapped into the PDO ...
  while (MappingSubidx <= NumObjsInPDO) // Note: This loop may run ZERO times !
     MappedObject := sdo.r( (CommPar+0x200).(MappingSubidx), U32, 0);
     NumBitsMapped := NumBitsMapped + (MappedObject & 0xFF); // add a number of *BITS* (!)
     // See DS301 V4.2.0 pages 135 .. 136 : As too often, the CANopen spec
     // left doubts about how react properly :
     // > If the CANopen device receives a PDO that is having more data bytes 
     // > than the number of mapped data bytes is (length), then the CANopen 
     // > device shall use the first data bytes up to the length and may be 
     // > initiate the EMCY write service, if supported.
     // "And maybe, or maybe not"... are you kidding ?
     // Na was denn nun, liebe CANopen Gurus ? Normal weitermachen und die 
     // "berzhligen Bytes" im PDO ignorieren, oder per EMERGENCY-Message
     // um Hilfe schreien ?
     // Oder um Hilfe schreien aber trotzdem den "zu langen PDO" verarbeiten ?
     MappingSubidx := MappingSubidx+1;
  endwhile;
  NumBytesMapped[CommPar] := int( (NumBitsMapped + 7 ) / 8 );
     // ,--------|_____|
     // '--> Note the use of the PDO's "communication parameter index",
     //      e.g. 0x1400, 0x1401, ... for RPDO (Receive-PDOs),
     //           0x1800, 0x1801, ... for TPDO (Transmit-PDOs) !
     // This is possible because in the CAN TESTER SCRIPT LANGUAGE,
     // array-indices are not necessarily zero-based as in "C", but stored
     // in the "array" itself (so the 'indices' may be arbitrary values
     // as long as they are unique, and fit inside a 32-bit integer) .
return; 



RestoreDefaults::
ClearScreen();
print( "\nTrying to restore the DEFAULT SETTINGS, the CANopen-way ..." );
print( "'Loading' default parameters via 0x1011" );
sdo.w( 0x1011.1, U32, 0x64616F6C) ; "load" DEFAULT PARAMETERS (!)  <<<<<
  ; The default values (after "load") become effective after a NODE RESET, not immediately !
  ; Do *NOT* call "save all parameters" after "load default parameters" !
  ; Actually, object 0x1011 doesn't "load" anything but "delete" the signature
  ; of e.g. the EEPROM storage, so ON THE NEXT RESET, the device is forced to use defaults
  ; (because there is nothing valid to "load" from EEPROM. CANopen's "load" is a misnomer.
print( "Node Reset (NMT) and wait for re-boot" );
000 81 00  ; Node Reset (required after "restore default parameters", don't ask why - it's CANopen ! )
delay(0.5)
print( "'Saving' default parameters via 0x1010" );
sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)      <<<<<
delay(0.5)
print( "Node Reset (NMT) and wait for re-boot" );
000 81 00  ; Node Reset (required after "load default parameters", don't ask why - it's CANopen ! )
delay(0.5)
print( "Entering 'Operational' state (NMT)" );
000 01 00  ; enter OPERATIONAL state, start PDOs. Only a TRANSITION sends all PDOs with tr-type 255
delay(0.5)
goto OtherOptions; // show the PDO-communication and -mapping parameters again

;----------------------------------------------------------------------
; Individueller Test der digitalen Ein/Ausgnge, fr diverse E/A-Module
;----------------------------------------------------------------------
  ; Befehle "auerhalb der Reihe" (vor Label "DigIoTest") :
  ; Beim Versuch, NICHT EXISTIERENDE Objekte anzusprechen,
  ;  muss der SDO-Server (im E/A-Modul) einen Fehler melden,
  ;  z.B. mit 'abort code' 11 00 09 06 = "subindex does not exist" !
  sdo.w(0x6200.1, U8, 0xFF ); // DO  1 ..  8  HIGH
  sdo.w(0x6200.2, U8, 0xFF ); // DO  9 .. 16  HIGH
  sdo.w(0x6200.3, U8, 0xFF ); // DO 17 .. 24  HIGH
  sdo.w(0x6200.4, U8, 0xFF ); // DO 25 .. 32  HIGH
  sdo.w(0x6200.5, U8, 0xFF ); // DO 33 .. 40  HIGH
  sdo.w(0x6200.6, U8, 0xFF ); // DO 41 .. 48  HIGH
  sdo.w(0x6200.7, U8, 0xFF ); // DO 49 .. 56  HIGH
  sdo.w(0x6200.8, U8, 0xFF ); // DO 57 .. 64  HIGH
  sdo.w(0x6200.1, U8, 0x00 ); // DO  1 ..  8  LOW
  sdo.w(0x6200.2, U8, 0x00 ); // DO  9 .. 16  LOW
  sdo.w(0x6200.3, U8, 0x00 ); // DO 17 .. 24  LOW
  sdo.w(0x6200.4, U8, 0x00 ); // DO 25 .. 32  LOW
  sdo.w(0x6200.5, U8, 0x00 ); // DO 33 .. 40  LOW
  sdo.w(0x6200.6, U8, 0x00 ); // DO 41 .. 48  LOW
  sdo.w(0x6200.7, U8, 0x00 ); // DO 49 .. 56  LOW
  sdo.w(0x6200.8, U8, 0x00 ); // DO 57 .. 64  LOW


SimpleLL:
  sdo.w($6300.1, U16, 0x0001 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #1 gesetzt
  sdo.w($6300.1, U16, 0x0002 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #2 gesetzt
  sdo.w($6300.1, U16, 0x0004 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #3 gesetzt
  sdo.w($6300.1, U16, 0x0008 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #4 gesetzt
  sdo.w($6300.1, U16, 0x0010 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #5 gesetzt
  sdo.w($6300.1, U16, 0x0020 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #6 gesetzt
  sdo.w($6300.1, U16, 0x0040 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #7 gesetzt
  sdo.w($6300.1, U16, 0x0080 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #8 gesetzt
  sdo.w($6300.1, U16, 0x0100 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #9 gesetzt
  sdo.w($6300.1, U16, 0x0200 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #10 gesetzt
  sdo.w($6300.1, U16, 0x0400 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #11 gesetzt
  sdo.w($6300.1, U16, 0x0800 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #12 gesetzt
  sdo.w($6300.1, U16, 0x1000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #13 gesetzt
  sdo.w($6300.1, U16, 0x2000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #14 gesetzt
  sdo.w($6300.1, U16, 0x4000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #15 gesetzt
  sdo.w($6300.1, U16, 0x8000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #16 gesetzt
  ; Seit 2017-01-19, fr das ZWEITE "16DO"-Modul am CANopen-Bus-Koppler :
  sdo.w($6300.2, U16, 0x0001 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #1 gesetzt
  sdo.w($6300.2, U16, 0x0002 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #2 gesetzt
  sdo.w($6300.2, U16, 0x0004 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #3 gesetzt
  sdo.w($6300.2, U16, 0x0008 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #4 gesetzt
  sdo.w($6300.2, U16, 0x0010 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #5 gesetzt
  sdo.w($6300.2, U16, 0x0020 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #6 gesetzt
  sdo.w($6300.2, U16, 0x0040 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #7 gesetzt
  sdo.w($6300.2, U16, 0x0080 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #8 gesetzt
  sdo.w($6300.2, U16, 0x0100 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #9 gesetzt
  sdo.w($6300.2, U16, 0x0200 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #10 gesetzt
  sdo.w($6300.2, U16, 0x0400 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #11 gesetzt
  sdo.w($6300.2, U16, 0x0800 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #12 gesetzt
  sdo.w($6300.2, U16, 0x1000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #13 gesetzt
  sdo.w($6300.2, U16, 0x2000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #14 gesetzt
  sdo.w($6300.2, U16, 0x4000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #15 gesetzt
  sdo.w($6300.2, U16, 0x8000 ); // Zugriff auf DO's als 16-Bit-Alias, NUR #16 gesetzt

@goto SimpleLL
  print "Dig. Inputs  1..8 = $%02lX",sdo.r($6200.1); // .. und wieder zurcklesen..
  print "Dig. Inputs 9..16 = $%02lX",sdo.r($6200.2);


DigIoTest::
 @NDI := sdo.r($6000.0,U8,0) // Wie viele digitale Inputs hat das Ding ?
 @NDO := sdo.r($6200.0,U8,0) // .. und wie viele digitale OUTPUTS ?
 if( (NDI<=0) && (NDO<=0) )  // Test ueberspringen wenn's weder DI noch DO gibt !
   goto FreqIOTest
 endif
 OpenWindow( "E/A-Modul-Test Schritt 2 : Digital-Ein/Ausgabe", w=640, h=720 )
 SetButton( 1, "<< &Neustart", goto Neustart  )
 SetButton( 2, "&Beenden",     goto Beenden   )
 SetButton( 3,"Alle &EIN", @R=0 : ctx((P1) [NumBytesMapped[0x1400]] 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF) : ctx((P2) 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF) ; per PDO alles EIN
 SetButton( 4,"Alle &AUS", @R=0 : ctx((P1) [NumBytesMapped[0x1400]] 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00) : ctx((P2) 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00) ; per PDO alles AUS
 SetButton( 5,"&Lauflicht",@R=1 )  ; Rampen-Flag wieder setzen (hier: Lauflicht)
 SetButton( 6, "&Weiter >>",   goto FreqIOTest)
 ClearScreen() : SetFont("Arial",20,1) : print("Digitale Ein/Ausgnge EA-xxx")
DigInfo:
 SetFont("Arial",12,1+2+4) : MoveTo(8,600) 
 print " Ausgnge auch mal KURZSCHLIESSEN - rote LED muss leuchten bzw blinken ! ")
 SetFont("Courier New",9) : MoveTo(10,40)
 print "Anzahl digitaler Eingnge      = %ld * 8bit", NDI
 if (NDI>=1)     // only if digital inputs exist...
   print "Ursprnglicher Zustand d.d.E.  = $%02lX",sdo.r($6000.1)
 endif
 print "Anzahl digitaler Ausgnge      = %ld * 8bit", NDO
 if (NDO>=1)     // only if digital outputs exist...
   print "Ursprnglicher Zustand d.d.A.  = $%02lX",sdo.r($6200.1)
 endif 
 if(NDO>=1) print "Port-Richtungen Gruppe 1 = %08b",sdo.r($5FF5.1)
 if(NDO>=2) print "Port-Richtungen Gruppe 2 = %08b",sdo.r($5FF5.2)
 gosub Bootup ; ehemals "Extended Bootup" fuer Node-Guarding, funzt mit SYS TEC nicht
 CommPar := 0x1400;  // find the number of bytes to send into the device's first RPDO..
 gosub GetSizeOfPDO; // in : CommPar,  out : NumBytesMapped[CommPar]
 CommPar := 0x1401;  // find the number of bytes to send into the device's 2nd RPDO..
 gosub GetSizeOfPDO; // in : CommPar,  out : NumBytesMapped[CommPar]
 000 01 00  ; Modul von Pre-Operational -> Operational schalten, PDOs aktiv.
     ; Alle PDOs mit transmission-type 255 nach DS401 sollten nun MINDESTENS EINMAL
     ;  gesendet werden, und darberhinaus auch noch EREGNISGESTEUERT .
 @N=0       ; Testmuster fuer digitale Outputs
 @R=1       ; Rampe (hier: Lauflich) per Default aktiv
 @tx_interval=0.005  // new TX interval, also limits execution speed: SECONDS per line
 if( (NDI+NDO) <= 16 )
     SetFont("Courier New",19)
 else if ( (NDI+NDO) <= 22)
     SetFont("Courier New",15)
 else
     SetFont("Courier New",12)
 endif
DigIoLoop:  ; digital-I/O per PDO ansteuern und per SDO zuruecklesen :
  D := table[N%8](0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80) ; Lauflichtmuster
  if(R==1)
   (P1) [NumBytesMapped[0x1400]] (D) (D) (D) (D) (D) (D) (D) (D)  ; bis zu 64 digitale Outputs PER PDO ansteuern
   (P2) [NumBytesMapped[0x1401]] (D) (D) (D) (D) (D) (D) (D) (D)  ; weitere 64 digitale (ODER ANALOGE) Outputs ansteuern
   0x080 // send SYNC just in case someone set the "PDO transmission type(s)" to 'SYNC' .. 
  endif
  MoveTo(0,120) 
  if(NDI>=1) print "Digitale Inputs  8 bis  1 = %08bbin ",sdo.r($6000.1)
  if(NDI>=2) print "Digitale Inputs 16 bis  9 = %08bbin ",sdo.r($6000.2)
  if(NDI>=3) print "Digitale Inputs 24 bis 17 = %08bbin ",sdo.r($6000.3)
  if(NDI>=4) print "Digitale Inputs 32 bis 25 = %08bbin ",sdo.r($6000.4)
  if(NDI>=5) print "Digitale Inputs 40 bis 31 = %08bbin ",sdo.r($6000.5)
  if(NDI>=6) print "Digitale Inputs 48 bis 39 = %08bbin ",sdo.r($6000.6)
  if(NDI>=7) print "Digitale Inputs 56 bis 49 = %08bbin ",sdo.r($6000.7)
  if(NDI>=8) print "Digitale Inputs 64 bis 57 = %08bbin ",sdo.r($6000.8)

  if(NDO>=1) print "Digitale Outputs  8 bis 1 = %08bbin ",sdo.r($6200.1)
  if(NDO>=2) print "Digitale Outputs 16 bis 9 = %08bbin ",sdo.r($6200.2)
  if(NDO>=3) print "Digitale Outputs 24 .. 17 = %08bbin ",sdo.r($6200.3)
  if(NDO>=4) print "Digitale Outputs 32 .. 25 = %08bbin ",sdo.r($6200.4)
  if(NDO>=5) print "Digitale Outputs 40 .. 31 = %08bbin ",sdo.r($6200.5)
  if(NDO>=6) print "Digitale Outputs 48 .. 41 = %08bbin ",sdo.r($6200.6)
  if(NDO>=7) print "Digitale Outputs 56 .. 49 = %08bbin ",sdo.r($6200.7)
  if(NDO>=8) print "Digitale Outputs 64 .. 57 = %08bbin ",sdo.r($6200.8)
  if(NDO>=9) print "Digitale Outputs 72 .. 65 = %08bbin ",sdo.r($6200.9)
  if(NDO>=10) print "Digitale Outputs 80 .. 71 = %08bbin ",sdo.r($6200.10)
  if(NDO>=11) print "Digitale Outputs 88 .. 81 = %08bbin ",sdo.r($6200.11)
  if(NDO>=12) print "Digitale Outputs 96 .. 89 = %08bbin ",sdo.r($6200.12)
  if(NDO>=13) print "Digitale Outp. 104 ..  97 = %08bbin ",sdo.r($6200.13)
  if(NDO>=14) print "Digitale Outp. 112 .. 103 = %08bbin ",sdo.r($6200.14)
  if(NDO>=15) print "Digitale Outp. 120 .. 113 = %08bbin ",sdo.r($6200.15)
  if(NDO>=16) print "Digitale Outp. 128 .. 121 = %08bbin ",sdo.r($6200.16)
  if(NDO>=17) print "Digitale Outp. 136 .. 129 = %08bbin ",sdo.r($6200.17)
  if(NDO>=18) print "Digitale Outp. 144 .. 137 = %08bbin ",sdo.r($6200.18)
  if(NDO>=19) print "Digitale Outp. 152 .. 143 = %08bbin ",sdo.r($6200.19)
  if(NDO>=20) print "Digitale Outp. 160 .. 153 = %08bbin ",sdo.r($6200.20)
  N := (N+1)%256
  delay(0.01)
@goto DigIoLoop


Bootup:    ; Unterprogramm "Bootup MIT Nodeguarding und SYNC" (ehemals Extended, aber..)
000 80 00  ; Enter Pre-Operational State
    ; Keine Atempause, Geschichte wird gemacht, es geht voran .. kein Warten auf Bootup !
000 01 00  ; Pre-Operational -> Operational, PDOs aktiv.
080 ; zu Testzwecken schon das erste 'SYNC' raushauen :) - ebenfalls OHNE PAUSE !
    ; (ein "guter Master" verhlt sich anders. Dies ist eher ein STRESS-TEST)
return  ; Ende von "Bootup"

;----------------------------------------------------------------------
; Individueller Test der ANALOGEN Ausgnge, u.A. fr "IO16 8AOV"
;----------------------------------------------------------------------
print("Dieses Modul hat", sdo.r($6411.0,U8,0), "analoge Ausgnge.");
AoutVal := 0;      // -> 0 Volt am Analogausgang ?
AoutVal := -32768; // -> -10 V beim "IO16 8AOV"
AoutVal := +32767; // -> +10 V beim "IO16 8AOV", 20 mA bei "4AOC" (Maximum)
AoutVal := +29490; // -> +9.000 V beim "IO16 8AOV"
AoutVal := +33;    // -> + 10.07 mV beim "IO16 8AOV" ?
AoutVal := -33;    // -> - 10.07 mV beim "IO16 8AOV" ?
AoutVal := +66;    // -> + 20.14 mV beim "IO16 8AOV" ?
AoutVal := -66;    // -> - 20.14 mV beim "IO16 8AOV" ?
AoutVal := +99;    // -> + 30.21 mV beim "IO16 8AOV" ?
AoutVal := -99;    // -> - 30.21 mV beim "IO16 8AOV" ?
AoutVal := 0;

;----------------------------------------------------------------------
AnalogOutRamp:: // <- another "top-level" label, listed in the menu unter 'Run'..'Run from Label'
  tx_interval := 0.001;  // execution speed limit : 1 ms per line
  freeze;  // pause RX/TX-display ("all CPU power for the loop below")
AnalogOutLoop:
  sdo.w($6411.1, I16, AoutVal ) : sdo.w($6411.2, I16, AoutVal ) :  sdo.w($6411.3, I16, AoutVal ) : sdo.w($6411.4, I16, AoutVal ); // set analog outputs 1..4
  sdo.w($6411.5, I16, AoutVal ) : sdo.w($6411.6, I16, AoutVal ) :  sdo.w($6411.7, I16, AoutVal ) : sdo.w($6411.8, I16, AoutVal ); // set analog outputs 5..8
  AoutVal := AoutVal + 200;
  if(AoutVal>32767)  AoutVal := -32768;
@goto AnalogOutLoop
print("Analog-Out 1 :", sdo.r($6411.1,I16) ) ;
print("Analog-Out 2 :", sdo.r($6411.2,I16) ) ;
print("Analog-Out 3 :", sdo.r($6411.3,I16) ) ;
print("Analog-Out 4 :", sdo.r($6411.4,I16) ) ;
print("Analog-Out 5 :", sdo.r($6411.5,I16) ) ;
print("Analog-Out 6 :", sdo.r($6411.6,I16) ) ;
print("Analog-Out 7 :", sdo.r($6411.7,I16) ) ;
print("Analog-Out 8 :", sdo.r($6411.8,I16) ) ;


; Test des 10-mV-"dead bands" (siehe AD567x-Datenblatt Seite 4, Funote zur Linearitt)
sdo.w($6411.1, I16, 0  ); // Theorie: 0.0 V   Gemessen: +2.5 mV
sdo.w($6411.1, I16, 16 ); // Theorie: 4.8 mV  Gemessen: +3.1 mV
sdo.w($6411.1, I16, 32 ); // Theorie: 9.8 mV  Gemessen: +7.6 mV

; Test des Filters fr ANALOGE AUSGNGE (Zeitkonstante per Oszilloskop messen) .
; Gemessene Zeitkonstante 'tau' des Ausgangsfilters beim "IO16 8AOV" : ca. 337 Mikrosekunden .
; Messungen und Details in C:\cproj\IO16_System\AD56xx\WB_Notizen_zum_AD567x_und_Messungen_am_IO16_8AOV.txt .
sdo.w($6411.1, I16, +32767 ); // Sprung von MAXIMUM (+10 V) ...
sdo.w($6411.1, I16, -32768 ); //    ... auf MINIMUM (-10 V)
sdo.w($6411.1, I16, 0      ); // Sprung auf NULL Volt
sdo.w($6411.1, I16, 8*3276 ); // Sprung auf +8 V
sdo.w($6411.1, I16, 5*3276 ); // Sprung auf +5 V
sdo.w($6411.1, I16,-5*3276 ); // Sprung auf -5 V

AnalogOutSteps:: ; "Heftiges Rappeln" an einem Analogausgang ( kein 'Overload' auch bei kapazitiver Last ? )
  sdo.w($6411.1, I16, +32767 ) : delay(0.005) 
  sdo.w($6411.1, I16, -32768 ) : delay(0.005) 
  sdo.w($6411.1, I16, +32767 ) : delay(0.005) 
  sdo.w($6411.1, I16, -32768 ) : delay(0.005)
goto AnalogOutSteps



;----------------------------------------------------------------------
ReadBackAnalogOutputs:: // <- another "top-level" label, listed in the menu unter 'Run'..'Run from Label'
// Relativ LANGSAM, weil die folgenden SDO-Lesezugriffe erst per IBUS durchgereicht werden mssen:
@NAO := sdo.r($5411.0,U8,0) // how many ANALOG OUTPUTS that can be "read back" ?
if(NAO>=1)
   Index := 0;
   repeat
      Index := Index + 1;
      print("READ BACK Analog-Out",Index,":", sdo.r($5411.(Index),I16) ) ;
   until Index >= NAO;
endif;
stop;


;----------------------------------------------------------------------
ReadModuleErrorFlags:: // <- another "top-level" label, listed in the menu unter 'Run'..'Run from Label'
@tx_interval=0.01  // set TX interval, also limits execution speed: SECONDS per line
// Auslesen der "Error-Flags" vom Buskoppler und allen Slave-Modulen. Aus CANopen-Sicht ein "Array".
// Jedes gesetzte Bit reprsentiert EINE der mglichen Fehlerquellen - Details im Gertehandbuch .
// Im Idealfall sollten alle Fehler-Flags NULL sein.
@NModules := sdo.r($5FE8.0,U8,0) // how many elements are in the array of 'error flags' ?
if(NModules>=1)
   Index := 0;
   repeat
      Index    := Index + 1;
      ErrFlags := sdo.r($5FE8.(Index), U32 );
      print("Module Error Flags [%ld] : %32b (binr)",Index,ErrFlags ) ;
      if( ErrFlags & 0x0001 ) 
        print( " Bit 0 : I2C bus timeout !" );
      endif;
      if( ErrFlags & 0x0002 ) 
        print( " Bit 1 : EEPROM read error !" );
      endif;
      if( ErrFlags & 0x0004 ) 
        print( " Bit 2 : EEPROM write error !" );
      endif;
      if( ErrFlags & 0x0008 ) 
        print( " Bit 3 : EEPROM checksum error !" );
      endif;
      if( ErrFlags & 0x0010 ) 
        print( " Bit 4 : Illegal EEPROM blocksize !" );
      endif;
      if( ErrFlags & 0x0020 ) 
        print( " Bit 5 : Modified slave module configuration !" );
      endif;
      if( ErrFlags & 0x0040 ) 
        print( " Bit 6 : Latched IBUS error !" );
      endif;
      if( ErrFlags & 0x0080 ) 
        print( " Bit 7 : Temporary IBUS error !" );
      endif;
      if( ErrFlags & 0x0100 ) 
        print( " Bit 8 : Illegal temperature !" );
      endif;
      if( ErrFlags & 0x0200 ) 
        print( " Bit 9 : Illegal supply voltage !" );
      endif;
      if( ErrFlags & 0x0400 ) 
        print( " Bit 10 : Excessive supply current !" );
      endif;
      if( ErrFlags & 0x0800 ) 
        print( " Bit 11 : I/O overload (voltage or current) !" );
      endif;
      if( ErrFlags & 0x1000 ) 
        print( " Bit 12 : UART1 receive timeout !" );
      endif;
      if( ErrFlags & 0x2000 ) 
        print( " Bit 13 : UART2 receive timeout !" );
      endif;
      if( ErrFlags & 0x4000 ) 
        print( " Bit 14 : Packet discarded due to CRC error !" );
      endif;
      if( ErrFlags & 0x8000 ) 
        print( " Bit 15 : IBUS slave enumeration failed !" );
      endif;
   until Index >= NModules;
endif;
stop;


;----------------------------------------------------------------------
ReadIBusErrorCounters:: // <- another "top-level" label, listed in the menu unter 'Run'..'Run from Label'
@tx_interval=0.01  // set TX interval, also limits execution speed: SECONDS per line
// Auslesen der "IBUS-Fehler-Zhler" aus dem Buskoppler und allen Slave-Modulen. Aus CANopen-Sicht ein "Array".
// Jedes Slave-Modul hat ZWEI Eintrge im Array : Einen Fehlerzhler fr UART1, und einen fr UART2 :
//  [1]=Buskoppler, [2]=Slave1, UART1,   [3]=Slave1, UART2,  [4]=Slave2, UART1,  [5]=Slave2, UART2, etc.
// Mit etwas Glck lsst sich so die 'am strksten durch Burst gestrte Verbindungsstelle' erkennen .
NCounters := sdo.r($5FE9.0,U8,0) // how many elements are in the array of 'error flags' ?
if(NCounters>=1)
   Index := 0;
   repeat
      Index    := Index + 1;
      print("IBUS Error Counter [%ld] : %ld",Index, sdo.r($5FE9.(Index), U32 ) ) ;
   until Index >= NCounters;
endif;
stop;

;----------------------------------------------------------------------
ReadUARTNoiseCounters:: // <- another "top-level" label, listed in the menu unter 'Run'..'Run from Label'
@tx_interval=0.01  // set TX interval, also limits execution speed: SECONDS per line
// Auslesen der "UART-Noise-Zhler". Indikator fr Strungen beim EMV-BURST-Test. 
// Woran der UART im LPC824 'Rauschen' bzw Strungen auf der RXD-Leitung erkennt,
// steht im Datenblatt der CPU (NXP LPC82x "UM10800 Rev 1.2") auf Seite 196 .
// Jedes Slave-Modul hat ZWEI Eintrge im Array : Einen Zhler fr UART1, und einen fr UART2 :
//  [1]=Buskoppler, [2]=Slave1, UART1,   [3]=Slave1, UART2,  [4]=Slave2, UART1,  [5]=Slave2, UART2, etc.
// hnlichkeiten mit 'ReadIBusErrorCounters' sind kein Zufall .
NCounters := sdo.r($5FEA.0,U8,0) // how many elements are in the array of 'error flags' ?
if(NCounters>=1)
   Index := 0;
   repeat
      Index    := Index + 1;
      print("UART Noise Counter [%ld] : %ld",Index, sdo.r($5FEA.(Index), U32 ) ) ;
   until Index >= NCounters;
endif;
stop;






;----------------------------------------------------------------------
; Test der Frequenz-Ein/Ausgnge, z.B. bei EA-Modul "SAHM" / "IP65".
;  Aus Kompatibilittsgrnden auch beim "IO16"-Buskoppler vorhanden.
;----------------------------------------------------------------------

FreqIOTest:  
 @NFO := sdo.r($5C11.0,U8,0) // Wie viele Frequenzausgaenge gibt's hier ? (vermutlich KEINE)
 @NFI := sdo.r($5C00.0,U8,0) & 0x0001  // Gibt's Frequenzeingaenge (maximal 1, meistens 0)
 // Test ueberspringen wenn's keinen Frequenzeingang gibt :
 if( (NFO<=0) && (NFI<=0) )
   goto AnalogIoTest 
 endif
 OpenWindow( "E/A-IP65-Test Schritt 3 : Frequenz-Ein/Ausgnge", w=640, h=720 )
 SetButton( 1,"<< &Zurck", goto DigIoTest )
 SetButton( 2,"&Beenden",   goto Beenden )
 SetButton( 3,"&Weiter >>", goto AnalogIoTest);
 ClearScreen() : SetFont("Arial",20,1) : print("Frequenz-Ein/Ausgnge EA-IP65")
FInfo:
 SetFont("Arial",11,1+2+4) : MoveTo(8,300) 
 print "Evtl Frequenzausgang mit Frequenzeingang verbinden, DIP-Schalter S9  beachten !"
 SetFont("Arial",8) : MoveTo(10,40)
 print "Anzahl F-Eingnge = %lX", sdo.r($5C00.0))  ; -> 1=Frequenz-Input #1 vorhanden
 print "Anzahl Subindizes = %lX", sdo.r($5C01.0))  ; Hoechster Subindex dieses Objekts
 print "Torzeit Eingang 1   = %ld ms",sdo.r($5C01.1))   ; Torzeit in Millisekunden (Default: 10ms)
 print "Messimpulse Eing1 = %ld",sdo.r($5C01.2))   ; Anzahl Pulse innerhalb der Torzeit
 print "Zeitdifferenz Eing1 = %ld s ",sdo.r($5C01.3) ; Differenz erster bis letzter Puls (us)
 print "Frequenz Eingang1 = %ld Hz ",sdo.r($5C01.4) ; daraus berechnete Frequenz (16bit, Hz)
 MoveTo(210,40)
 print "Anzahl F-Ausgnge = %ld", NFO ; Wieviele Frequenzausgaenge ?
FLoop:
   MoveTo(0,150) : SetFont("Courier New",20)
   print( "Frequenzeingang  : %7.1f Hz   ", 0.1 * sdo.r($5C01.5))  ; 32 bit, Einheit= 0.1-Hz
   if(NFO>=1) print( "Frequenzausgang 1: %6ld  Hz   ",sdo.r($5C11.1))
   if(NFO>=2) print( "Frequenzausgang 2: %6ld  Hz   ",sdo.r($5C11.2))
   ; Neue Frequenzen ausgeben:
   if(NFO>=1) @sdo.w($5C11.1,U16, F)      ; Rechteckgenerator an F_out #1 aktivieren
   if(NFO>=2) @sdo.w($5C11.2,U16, 8000-F) ; Rechteckgenerator an F_out #2 aktivieren
   @F=(F+10)%8000
goto FLoop
stop  ; der folgende Abschnitt dient nur fr den EXPERTEN zur Fehlersuche :
sdo.w($5C01.1,U16, 100)   ; Torzeit fuer Frequenzzaehler setzen
sdo.w($5C01.1,U16, 10)    ; Torzeit fuer Frequenzzaehler setzen
@print "%ld frequency output(s)",sdo.r($5C11.0)) ; Wieviele Frequenzausgaenge ?
@print "FreqOut1_Hz=%ld",sdo.r($5C11.1))  ; AUSGANGS(!)-Frequenz (Hz) Kanal 1 ?
@sdo.w($5C11.1,U16, 11)  ; Ausgangsfrequenz fuer F_out #1 setzen
@sdo.w($5C11.2,U16, 13)  ; Ausgangsfrequenz fuer F_out #2 setzen
@sdo.w($5C11.1,U16, 0)    ; Rechteckgenerator an F_out #1 abschalten
@sdo.w($5C11.2,U16, 0)    ; Rechteckgenerator an F_out #2 abschalten
FLoop2:
  @sdo.w($5C11.1,U16, F)       ; Rechteckgenerator an F_out #1
  @sdo.w($5C11.2,U16, 8000-F)  ; Rechteckgenerator an F_out #2
  @F=(F+10)%8000
@goto FLoop2


;----------------------------------------------------------------------
; Test der analogen Ein/Ausgnge, fr diverse E/A-Module
;----------------------------------------------------------------------

AnalogIoTest:
 @NAI := sdo.r($6401.0,U8,0) // read number of ANALOG INPUTS (0 if access fails)
 @NAO := sdo.r($6411.0,U8,0) // read number of ANALOG OUTPUTS (0 if access fails)
 if( (NAI<=0) && (NAO<=0) )
   goto Fertig // Test ueberspringen wenn's weder AI noch AO gibt
 endif
 OpenWindow( "E/A-Modul-Test Schritt 4 : Analoge Ein/Ausgabe", w=640, h=720 )
 SetButton( 1,"<< &Zurck", goto DigIoTest )
 SetButton( 2,"&Beenden",   goto Beenden )
 if(NAO>0) SetButton( 3,"Alles auf M&in", sdo.w($6411.1,I16,-32768):sdo.w($6411.2,I16,-32768):sdo.w($6411.3,I16,-32768):sdo.w($6411.4,I16,-32768)  )
 if(NAO>0) SetButton( 4,"Alles auf &Null",sdo.w($6411.1,I16,0):sdo.w($6411.2,I16,0):sdo.w($6411.3,I16,0):sdo.w($6411.4,I16,0) )
 if(NAO>0) SetButton( 5,"Alles auf M&ax", sdo.w($6411.1,I16,32767):sdo.w($6411.2,I16,32767):sdo.w($6411.3,I16,32767):sdo.w($6411.4,I16,32767) )
 SetButton( 6,"&Weiter >>", goto Fertig );
 ClearScreen() : SetFont("Arial",20,1) : print("Analoge Ein/Ausgnge EA-xxx")
AInfo:
 SetFont("Arial",11,1+2+4) : MoveTo(8,610) 
 print "Analog-Aus- und -Eingnge verbinden, DIP-Schalter beachten !"
 SetFont("Arial",8) : MoveTo(10,40)
 print "Anzahl A-Eingnge =%ld", NAI
 print "Anzahl A-Ausgnge =%ld", NAO
 if(NAI>=1) print "Startzustand AIN1 =%ld",sdo.r($6401.1)
 if(NAI>=2) print "Startzustand AIN2 =%ld",sdo.r($6401.2)
ALoop:
   MoveTo(0,100) : SetFont("Courier New",20)
   if(NAO>=1)
      Index := 0;
      repeat
         Index := Index + 1;
         Aout[Index] := sdo.r($6411.(Index),I16);
         print( "Analogausgang %2d  : %6ld", Index, Aout[Index]; );
         ClearEOL(); // clear old garbage, clicker-free 
         print;  // new line AFTER clearing until the end-of-line (above)
      until Index >= NAO;
   endif;
   if(NAI>=1)
      Index := 0;
      repeat
         Index := Index + 1;
         Ain[Index] := sdo.r($6401.(Index),I16);
         print( "Analogeingang %2d  : %6ld", Index, Ain[Index]; );
         ClearEOL(); // clear old garbage, clicker-free 
         print;  // new line AFTER clearing until the end-of-line (above)
      until Index >= NAI;
   endif;
goto ALoop


;--------- Alle Tests durch... -----------------------------------------------------
Fertig:
 000 81 00      ; NMT-Telegramm: Operational -> Initialisation (d.h. Reset)
 OpenWindow( "CAN-Test : FERTIG ! ",w=320,h=200,bc=(NumErrors==0)?green:yellow)
 SetFont("Arial",18,2) : MoveTo(0,40) 
 if(NumErrors==0)
    @sound( 10, 600, 100 )
    print " Glckwunsch, das war's !"
 else
    @sound( 1200, 600, -100 )
    print " Naja, %ld Fehler..",NumErrors
 endif
 print "   .. und was nun ?"
 SetButton( 1,"&Neustart", goto Neustart )
 SetButton( 2,"&Beenden",  goto Beenden )
FertigW: ; "Fertig, Warte aber noch auf Button-Event 'Beenden' "
goto FertigW



;--------- Bei Fehlern... ----------------------------------------------------------
Verflixt:
 if(ErrorClass==ErrorClass.sdo)  ; // most likely a non-existing object, or SDO timeout ?
   NumErrors := NumErrors+1 : delay(0.5) : ResumeError // Weitermachen !
 endif
 OpenWindow( "CAN-Test : FEHLER ! ",w=400,h=200,bc=red)
 SetFont("Arial",8) : MoveTo(8,4) 
 print( "Fehlercode %ld (Klasse %d) in Programmzeile %ld", ErrorInfo.code, ErrorClass, ErrorInfo.line )
 print( ErrorInfo.text )
 SetFont("Arial",11,2) : MoveTo(8,40) 
 print "- CAN-Baudrate berprfen (momentan %ld kBit/sec)",can.baudrate
 print "- Knotennummer berprfen (momentan #%ld )",sdo.node
 print "- die verdammten Kabel berprfen"
 print "- die DIP-Schalter richtig einstellen"
 SetButton( 1,"&Nochmal versuchen",  goto Neustart )
 SetButton( 2,"&Abbrechen",  goto Abbruch )
 ClearError()   ; OnError-Handler wieder "scharf" machen
Fehlerschleife:
goto Fehlerschleife
Abbruch:  ; ABBRUCH im Fehlerfall, ohne nochmaligen CAN-Sendeversuch !
 CloseWindow()  ; Benutzerdefiniertes Fenster schliessen (falls noch offen)
 print "Programmabbruch - Neustart mit F9 !"
@stop



;--------- Test Ende ---------------------------------------------------------------
Beenden:  ; Planmaessiges TEST-ENDE, mit Abschalten der PDOs
 000 81 00      ; NMT-Telegramm: Operational -> Initialisation (d.h. Reset)
 CloseWindow()  ; Benutzerdefiniertes Fenster schliessen (falls noch offen)
 print "Programmende - Neustart mit F9 !"
@stop           ; stop command program. Press F9 to restart .



;----------------------------------------------------------------------
; FINGER WEG VOM FOLGENDEN ABSCHNITT !
; Der folgende Abschnitt dient nur fr den EXPERTEN zur Fehlersuche !
;----------------------------------------------------------------------
ExpertSettings::
@stop   ; STOP HERE --- some write accesses follow,
        ; only intended for the software developer
        ; using single-step execution of this command file !

@sdo.w($1011.1, U32, 0x64616F6C) ; "load" DEFAULT PARAMETERS (!)  <<<<<
 ; The default values (after "load") become effective after a NODE RESET, not immediately !
 ; Do *NOT* call "save all parameters" after "load default parameters" !
000 81 00  ; Node Reset (required after "load default parameters", don't ask why - it's CANopen ! )
delay(0.5)  ; simply WAIT until the CANopen-node has rebooted
@sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)      <<<<<
000 81 00  ; Node Reset (required after "load default parameters", don't ask why - it's CANopen ! )
delay(0.5) ; WAIT again until the CANopen-node has rebooted
000 01 00  ; enter OPERATIONAL state, start PDOs. Only TRANSITION sends all PDOs with tr-type 255 !

10000581.x r XX XX XX XX XX XX XX XX // request transmission of a MANFACTURER SPECIFIC PDO via RTR
   // (RTRs seem to be the only way to read analog inputs via PDO,
   //  due to the stupid DEFAULT settings from DS401 V3.0 page 103.
   //  Also, the TPDO's *EVENT TIMERS* must be off per default. Baaah. )
@sdo.w ($1801.5,U16,100) ; turn on TPDO2's "event timer" [periodic TX in ms, AIN  1.. 4]
@sdo.w ($1802.5,U16,100) ; turn on TPDO3's "event timer" [periodic TX in ms, AIN  5.. 8]
@sdo.w ($1803.5,U16,100) ; turn on TPDO4's "event timer" [periodic TX in ms, AIN  9..12]
@sdo.w ($1804.5,U16,100) ; turn on TPDO5's "event timer" [periodic TX in ms, AIN 13..16]
@sdo.w ($1805.5,U16,100) ; turn on TPDO6's "event timer" [periodic TX in ms, AIN 17..20]
@sdo.w ($1806.5,U16,100) ; turn on TPDO7's "event timer" [periodic TX in ms, AIN 21..24]
@sdo.w ($1807.5,U16,100) ; turn on TPDO8's "event timer" [periodic TX in ms, AIN 25..28]
@sdo.w ($1808.5,U16,100) ; turn on TPDO9's "event timer" [periodic TX in ms, AIN 29..32]






; Schreibzugriff auf 0x1000.00 DARF nicht funktionieren
; (das CANopen Conformance Test Tool versucht dies auch..):
@sdo.w ($1000.0,U32,0) ; illegal write to the "device type"

; SDO-Zugriff mit falschen 'command specifier' - auch dies wird beim 'Conformance Test' abgeklopft,
; und muss mit einem "Abort Code" 0x05040001 oder 0x08000000 beantwortet werden :
601.s E0 00 10 00 00 00 00 00

; SDO-Schreibzugriff mit "zu wenig" oder "zu vielen" Datenbytes:
; Conformance Test : "Write u32 to u16 or u8 object", "Write u8 to u16 or u32 object" :
; Um diesen Test erfolgreich zu bestehen, muss Objekt 0x1003 Subindex 0 auch SCHREIBBAR sein !
; DS301 V4.2.0 sagt dazu auf Seite 97 :
; > Writing 00h to sub-index 00h shall delete the entire error history (empties the array). 
; > Other values than 00h are not allowed and shall lead to an abort message (error code: 0609 0030h).
@sdo.w ($1003.0, U32, 0 );  // -> "Valid abort codes are 06070010h, 06070012h or 08000000h"
@sdo.w ($1005.0, U8,  0 );  // -> "Valid abort codes are 06070010h, 06070013h or 08000000h"
; Conformance Test : "Write 100Ch:00h segmented - if guarding is supported."
;  (d.h. ein SEGMENTED Transfer mit 2 Bytes, wo eigentlich ein EXPEDITED Transfer ausreichen wrde)
601.s 21 0C 10 00 02 00 00 00  ; -> Antwort: 581.s 60 0C 10 00 00 00 00 00
601.s 0B 00 00 00 00 00 00 00  ; zweiter Teil CLIENT -> SDO-Server: "Zwei von Zwei Bytes"
@sdo.w ($100C.0, U32, 0 );

; Schreibzugriffe auf diverse "COB-IDs" mit den in DS301 aufgelisteten "Restricted IDs"
; mssen mit einem bestimmten "Abort Code" abgewiesen werden:
@sdo.w ($1005.0,U32,0x180) ; illegal write a 'restricted ID' to "COB-ID Sync Message"



print "CobID_Sync=$%08lX",sdo.r($1005.0)
@sdo.w ($1005.0,U32,$40000080) ; configure SYNC-Object. Bit 30 is the option to SEND Sync.
@print "CommCycle=%ld ms",sdo.r($1006.0)/1000 ; "Communication Cylce" aka "SYNC TX Cycle"
@sdo.w ($1006.0,U32,20000)  ; set comm-cycle to NNN microseconds (misleading name for SYNC CYCLE)
; Surprising but true: SYNC messages must already be sent in PRE-OPERATIONAL state ! !
; To TURN OFF the transmission of SYNC messages, clear bit 30 in object 0x1005 ...
@sdo.w ($1005.0,U32,$00000080) ; configure SYNC-Object. Bit 30 cleared means don't SEND but RECEIVE.
; or, to turn off the transmission of SYNC messages, set the "comm-cycle" to ZERO:
@sdo.w ($1006.0,U32,0) ; setting the comm-cycle to zero STOPS sync
; Besides the above simple "Sync" configuration objects, there's a strange thing
; that CANopen DS301 calls "Synchronous counter overflow value" (object 0x1019):
; > If the value is greater than 1, the SYNC message shall have a data length of 1 byte. 
; > The first data byte contains the counter.
@print "SyncCounterOverflowValue=%ld",sdo.r($1019) ; "Communication Cylce" aka "SYNC TX Cycle"
; Note : MODIFYING the "Synchronous counter overflow value" is only allowed
;        while the "Sync Cycle", h.. sorry.. "Communication Cycle" is ZERO !
@sdo.w ($1019.0,U8,0) ; "Synchronous counter overflow value" = 0 -> send SYNC *without* any data
@sdo.w ($1019.0,U8,1) ; "Synchronous counter overflow value" = 1 -> illegal, must return an SDO ABORT
@sdo.w ($1019.0,U8,2) ; "Synchronous counter overflow value" = 2 -> send SYNC with a 1-byte counter, HIGHEST VALUE = 2
@sdo.w ($1019.0,U8,240) ; maximum allowed "Synchronous counter overflow value"
@sdo.w ($1019.0,U8,241) ; ILLEGAL "Synchronous counter overflow value" (Mr Pedantic wants to see an ABORT CODE now)







000 01 00  ; Pre-Operational -> Operational, start PDO 
@print sdo.r($1012.254)     ; Name of this object ?
@print "TimeStampObj=$%08lX",sdo.r($1012.0)
@print sdo.r($1014.254)     ; Name of this object ?
@print "EmergcyObject=$%08lX",sdo.r($1014.0)
@print sdo.r($1015.254)     ; Name of this object ?
@print "EmergencyInhib=%ld",sdo.r($1015.0)
@print sdo.r($1016.254)     ; Name of this object ?
@print "HeartbeatConsumers=%ld",sdo.r($1016.0)
@print "HeartbeatConsumer1=$%08lX",sdo.r($1016.1)
@print "HeartbeatConsumer2=$%08lX",sdo.r($1016.2)
@print sdo.r($1017.254)     ; Name of this object ?
@print "HeartbeatProducerTime=%ld ms",sdo.r($1017.0)
@print sdo.r($1018.254)     ; Name of this object ?
@print "Vendor ID=$%08lX",sdo.r($1018.1) ; MKT=0x004D4B54
@print "Prod.Code=$%08lX",sdo.r($1018.2)
@print "RevisionN=$%08lX",sdo.r($1018.3)
@print "SerialNr.=$%08lX",sdo.r($1018.4)
  ; ex: @print "DigOutputMaskA = %08b",sdo.r($6208.1,U8,0) // Filter Mask Digital Output 8-bit
  ; DS401-spezifischer Quark in einem der folgenden Abschnitte !

ReadEDSFiles::
print "Trying to read EDS file(s), turning off the CAN message display.."
freeze := TRUE; // stop CAN message display to speed up reading EDS-file via SDO

print sdo.r($1021.0,str) ; Try to read the 'stored EDS' from this device 
         ; (CANopen DS301 says the data type is 'DOMAIN', in fact it's plain text .
         ;  Because it's LARGE, the CAN-tester stores the read-out data as a file )
print "Alternative EDS formats: %d",sdo.r($5021.0,U8) ; Do we have "alternative" formats for the EDS file ?
         ; 2017-06-26 : 0x5021.1 = "large EDS", 0x5021.2 = "small EDS", 0x5021.3 = "compact EDS", 0x5021.4 = DCF .
print sdo.r($5021.1,str) ; generates the normal, 'large' EDS (> 120 kByte result stored as a file),
                         ; Passes the CANopen Conformance 'OD' tests, but not ok for Codesys.
print sdo.r($5021.2,str) ; generates a 'smaller' EDS for Codesys. Omits thousands of SUB-OBJECTS.
                         ; Ok for Codesys, but miserably fails most CANopen Conformance 'OD' tests.
print sdo.r($5021.3,str) ; generates an EDS with 'compact storage', which didn't pass the "Conformance Test"
                         ; because "compact sub-objects" with a VARIABLE array size seem to be impossible.
print sdo.r($5021.4,str) ; doesn't generate an EDS but a DCF (device configuration file)
freeze := FALSE;  // let CAN message displays run again (RX, TX)
stop;




; Try to reproduce what when wrong in the "conformance test", "OD 02" :
;  > Compare the read value against the default value in the EDS if it exists.
@sdo.w($1011.1, U32, 0x64616F6C) ; "load" DEFAULT PARAMETERS (!) 
000 81 00 ; Node Reset (restores the "default values" with a bit of luck)
@print "SDO_Server.Sub0= %ld",sdo.r($1200.0)     ; expected : 2 (highest sub-index)
@print "SDO_Server.C_S = 0x%08lX",sdo.r($1200.1) ; expected : 0x600 + NodeID
@print "SDO_Server.S_S = 0x%08lX",sdo.r($1200.2) ; expected : 0x580 + NodeID
  ; (unlike any other SDO-server, the above "first" SDO-server only has TWO
  ;  sub-indices, thus the default value of TWO, not THREE, in subindex 0 )



; Try to reproduce what when wrong in the "conformance test", "PDO 23" :
000 81 00 ; Node Reset
; # Test case : 3.23  -  Test name : PDO 23 ####################################
; Test Info : Try to write a new correct mapping without the mapping procedure}(node-ID)
print "RPDO1 COB-ID=%08lX", sdo.r($1400.1, U32) 
print "RPDO1 num mapped objects=%d", sdo.r($1600.0, U8) 
sdo.w($1400.1, U32, 0x80000201 ) ; CANopen DS301 : "destroy PDO" (make it invalid via COMM-PAR.)
sdo.w($1600.1, U32, 0x62000108 ) ; try to change mapping without the mandatory "mapping procedure"
sdo.w($1400.1, U32, 0x00000201 ) ; "undestroy PDO" (by clearing bit 31 in the "COB ID")



; Try to reproduce what when wrong in the "conformance test", "PDO 24" .
;      First the CORRECT sequence (to set a PDO-mapping-objet) ...
print "TPDO1 COB-ID=%08lX", sdo.r($1800.1, U32) 
print "TPDO1 num mapped objects=%d", sdo.r($1A00.0, U8) 
sdo.w($1800.1, U32, 0x80000181 ) ; CANopen DS301 : "destroy PDO".  WB: What a misleading phrase !
sdo.w($1A00.0, U8, 0 )
sdo.w($1A00.1, U32, $60000108 ) ; try to change mapping without the mandatory "mapping procedure"
sdo.w($1A00.0, U8, 1 )
sdo.w($1800.1, U32, 0x00000181 ) ; "undestroy PDO" (by clearing bit 31 in the "COB ID")


; Try to reproduce what when wrong in the "conformance test", "States - NMT Slave Machine" ...
;  "Test Error : State 01: Accessing object 1400h:03h - current value (0h) - default value (FFh) Mismatch!
;  (A FEW objects were not set back to their 'default values' during a Node Reset .
;   Full story in C:\cproj\IO16_System\EDS\WB_Notizen_zum_CANopen_Conformance_Test.txt )
@sdo.w($1011.1, U32, 0x64616F6C) ; "load" DEFAULT PARAMETERS (!) 
000 81 00                        ; Node Reset (required after "load default parameters")
print "RPDO1 Inhibit Time=0x%04lX", sdo.r($1400.3, U16) ; RPDO1 inhibit time == default value in EDS file ?
@sdo.w($1400.3, U16, 0x1234)     ; *temporarily* modify RPDO1 inhibit time
print "RPDO1 Inhibit Time=0x%04lX", sdo.r($1400.3, U16) ; new RPDO1 inhibit time really set ?
000 81 00                        ; Node Reset (the value set above must get LOST now)
print "RPDO1 Inhibit Time=0x%04lX", sdo.r($1400.3, U16) ; RPDO1 inhibit time back to the 'default' value ?



print "OS-Interface: %d subs", sdo.r($1023.0,U8) ; Does an 'OS command interface' exist ? How many subindices (expected: 3)
; Can we retrieve DATA TYPE DEFINITIONS via SDO ? ("DEFTYPES", purpose = "TYPEDEFS", harrr, harrr...)
;    See notes in DS301 V3.2 page 90..91 about the purpose of these OD entries.
print sdo.r($0001.0,U32) ; Data type 0x01 = BOOLEAN. May not be supported -> (object doesn't exist)
print sdo.r($0002.0,U32) ; Data type 0x02 = INTEGER8   : expected result =  8 [bits]
print sdo.r($0002.1,U32) ; Data type 0x02 = INTEGER8   : expected result =  SUBINDEX DOES NOT EXIST (because a BYTE is not a 'complex' type)
print sdo.r($0003.0,U32) ; Data type 0x03 = INTEGER16  : expected result = 16 [bits] 
print sdo.r($0004.0,U32) ; Data type 0x04 = INTEGER32  : expected result = 32 [bits] 
print sdo.r($0005.0,U32) ; Data type 0x05 = UNSIGNED8  : expected result =  8 [bits]
print sdo.r($0006.0,U32) ; Data type 0x06 = UNSIGNED16 : expected result = 16 [bits]
print sdo.r($0007.0,U32) ; Data type 0x07 = UNSIGNED32 : expected result = 32 [bits]
print sdo.r($0008.0,U32) ; Data type 0x08 = REAL32     : expected result = 32 [bits]
print sdo.r($0009.0,U32) ; Data type 0x09 = VIS_STRING : expected result = 0  [bits] because the length is VARIABLE (and thus UNKNOWN here)
print sdo.r($000A.0,U32) ; Data type 0x0A = OCT_STRING : expected result = 0  [bits] '''
print sdo.r($000C.0,U32) ; Data type 0x0C = UNICODE_STR: expected result = 0  [bits] '''
print sdo.r($000C.0,U32) ; Data type 0x0B = TIME_OF_DAY: expected result = 48 [bits]
print sdo.r($000D.0,U32) ; Data type 0x0B = TIME_DIFF  : expected result = 48 [bits]




;-------------- "transmit" - PDO (analoge und digitale Inputs) -------------
   ;
   ; 0x1800 = 1st TPDO Communication parameter
@print "TxPDO1CommPar.Sub0=%ld",sdo.r($1800.0)
@print "TxPDO1CommPar.CobID=$%08lX",sdo.r($1800.1) ; PDO ID of 1st TPDO    ?
@sdo.w ($1800.1, U32, 0x00000184)                  ; set TPDO1 CAN id
 ; In a strict CANopen implementation, this should usually FAIL because
 ;   DS301 V4.02 says:  "It is not allowed to change bit 0..29 
 ;   while the PDO exists." No further explanation by CiA; 
 ;   SYS TEC correctly returnED error "object is READ-ONLY" in that case.
 ; But CoDeSys didn't care, so WoBu was forced to modify this behaviour !
 ; More info: Grep for "PDO_KLUDGE_1_FOR_CODESYS" in the canopen source .
@print "TxPDO1CommPar.TrTyp= %03ld",sdo.r($1800.2) ; PDO transmission type ?
 ; (tr.type: 255=event  1..240=sync,  253=on RTR    [DS301 10-6] )
@sdo.w ($1800.2, U8, 255)                          ; set PDO transmission type
@print "TxPDO1CommPar.Inhib= %05ld",sdo.r($1800.3) ; PDO inhibit time      ?
@sdo.w ($1800.3, U16, 123)                         ; set PDO inhibit time
@print "TxPDO1CommPar.CMSpr= %03ld",sdo.r($1800.4) ; PDO CMS priority (no) ?
 ; <- Diesen Eintrag gibt's in CANopen "V4" nicht mehr, sollte 'nen Fehler geben:
 ;  80 00 18 04 11 00 09 06 = "subindex doesn't exist" .
@print "TxPDO1CommPar.EvTim= %05ld",sdo.r($1800.5) ; PDO event timer [ms]  ?
@sdo.w ($1800.5, U16, 1) ; set TPDO1 event timer [ms] (-> TXcycle if TxType==255)
@sdo.w ($1800.5, U16, 0) ; 0 = PDO event timer off

   ;
   ; 0x1A00 = 1st TPDO Mapping parameter
   ; Auch hier gab's einigen Trouble mit CoDeSys !
@print "TxPDO1Mapping.NrObj= %05ld",sdo.r($1A00.0) ; Number of mapped objs ?
@print "TxPDO1Mapping.Obj.1=$%08lX",sdo.r($1A00.1) ; First mapped object   ?
@print "TxPDO1Mapping.Obj.2=$%08lX",sdo.r($1A00.2) ; 2nd   mapped object   ?
@print "TxPDO1Mapping.Obj.3=$%08lX",sdo.r($1A00.3) ; 3rd   mapped object   ?
@print "TxPDO1Mapping.Obj.4=$%08lX",sdo.r($1A00.4) ; 4th   mapped object   ?
@stop  ; A "user-defined" PDO-Mapping for the 1st Transmit, for I/O-MODULE:
  ; In a STRICT implementation, this only works if PdoCommPar.ID.bit31=1  !
@sdo.w($1A00.0, U8,  0)  ; Begin to modify mapping table for TPDO1
@sdo.w($1A00.1, U32, 0x60000108) ; 1st mapped object: 8 digital inputs
@sdo.w($1A00.2, U32, 0x60000108) ; 2nd mapped object: the same once again !
@sdo.w($1A00.3, U32, 0x60000108) ; 3rd mapped object: the same once again !
@sdo.w($1A00.0, U8,  3)  ; Map X objects into TPDO1
@sdo.w($1010.1, U32, 0x65766173) ; SAVE ALL PARAMETERS (!)
@stop  ; A "user-defined" PDO-Mapping for the 1st Transmit, for UPT-167:
@sdo.w($1A00.0, U8,  0)  ; Begin to modify mapping table for TPDO1
@sdo.w($1A00.1, U32, 0x40010108) ; 1st mapped object: 8 digital inputs
@sdo.w($1A00.2, U32, 0x60000108) ; 2nd mapped object: the same once again !
@sdo.w($1A00.3, U32, 0x60000108) ; 3rd mapped object: the same once again !
@sdo.w($1A00.0, U8,  3)  ; Map X objects into TPDO1
@sdo.w($1010.1, U32, 0x65766173) ; SAVE ALL PARAMETERS (!)


   ;
   ; 0x1801 = 2nd TPDO Communication parameter
@print "TxPDO2CommPar.Sub0=%ld",sdo.r($1801.0)
@print "TxPDO2CommPar.CobID=$%08lX",sdo.r($1801.1) ; PDO ID of 2nd TPDO    ?
@sdo.w ($1801.1, U32, 0x80000284)  ; set TPDO2 CAN id (bit 31="disabled"!)
 ; See note on 'strict CANopen implementation' in the section for TPDO1 !
@print "TxPDO2CommPar.TrTyp= %03ld",sdo.r($1801.2) ; PDO transmission type ?
 ; (tr.type: 255=event 254=udef  1..240=sync,  253=on RTR    [DS301 10-6] )
@sdo.w ($1801.2, U8, 254)                          ; set PDO transmission type
@print "TxPDO2CommPar.Inhib= %05ld",sdo.r($1801.3) ; PDO inhibit time      ?
@sdo.w ($1801.3, U16, 10000) ; set PDO inhibit time to n * 0.1 millisec
@print "TxPDO2CommPar.EvTim= %05ld",sdo.r($1801.5) ; PDO event timer [ms]  ?
@sdo.w ($1801.5, U16, 1000) ; set TPDO2 event timer [ms] (-> TXcycle if TxType==255)
@sdo.w ($1801.5, U16, 50)   ; set TPDO2 event timer to 50 [ms]
@sdo.w ($1801.5, U16, 0)    ; 0 = PDO event timer off (no more "periodic but asynchronous" transmission)

   ;
   ; 0x1A01 = 2nd TPDO Mapping parameter
@print "TxPDO2Mapping.NrObj= %05ld",sdo.r($1A01.0) ; Number of mapped objs ?
@print "TxPDO2Mapping.Obj.1=$%08lX",sdo.r($1A01.1) ; First mapped object   ?
@print "TxPDO2Mapping.Obj.2=$%08lX",sdo.r($1A01.2) ; 2nd   mapped object   ?
@print "TxPDO2Mapping.Obj.3=$%08lX",sdo.r($1A01.3) ; 3rd   mapped object   ?
@print "TxPDO2Mapping.Obj.4=$%08lX",sdo.r($1A01.4) ; 4th   mapped object   ?


   ;
   ; 0x1802 = 3rd TPDO Communication parameter
@print "TxPDO3CommPar.Sub0=%ld",sdo.r($1802.0)
@print "TxPDO3CommPar.CobID=$%08lX",sdo.r($1802.1) ; PDO ID of 2nd TPDO    ?
@print "TxPDO3CommPar.TrTyp= %03ld",sdo.r($1802.2) ; PDO transmission type ?
 ; (tr.type: 255=event 254=udef  1..240=sync,  253=on RTR    [DS301 10-6] )
@sdo.w ($1802.2, U8, 255)                          ; set PDO transmission type
@print "TxPDO3CommPar.Inhib= %05ld",sdo.r($1802.3) ; PDO inhibit time      ?
@sdo.w ($1802.3, U16, 1000) ; set PDO inhibit time to n * 0.1 millisec
@print "TxPDO3CommPar.EvTim= %05ld",sdo.r($1802.5) ; PDO event timer [ms]  ?
@sdo.w ($1802.5, U16, 1000) ; set TPDO3 event timer [ms] (-> TXcycle if TxType==255)
@sdo.w ($1802.5, U16, 10)   ; set TPDO3 event timer to 10 [ms]
@sdo.w ($1802.5, U16, 0) ; 0 = PDO event timer off

   ;
   ; 0x1A02 = 3rd TPDO Mapping parameter
@print "TxPDO3Mapping.NrObj= %05ld",sdo.r($1A02.0) ; Number of mapped objs ?
@print "TxPDO3Mapping.Obj.1=$%08lX",sdo.r($1A02.1) ; First mapped object   ?
@print "TxPDO3Mapping.Obj.2=$%08lX",sdo.r($1A02.2) ; 2nd   mapped object   ?
@print "TxPDO3Mapping.Obj.3=$%08lX",sdo.r($1A02.3) ; 3rd   mapped object   ?
@print "TxPDO3Mapping.Obj.4=$%08lX",sdo.r($1A02.4) ; 4th   mapped object   ?

@stop  ;  Modify TPDO #3 ...
@sdo.w($1802.1, U32, 0x00000381) ; Modify CAN-Identifier used by PDO
@sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)
000 81 00  ; Node Reset (to test if REALLY saved)
000 01 00  ; Pre-Operational -> Operational, start PDO + SYNC 

@print "TPDO3_CobID=$%08lX",sdo.r($1802.1) ; read to check storage



;-------------- "receive" - PDO (analoge+digitale Outputs) -----------------
   ;
   ; 0x1400 = 1st RPDO Communication parameter
print "RxPDO1CommPar.Sub0=%ld",sdo.r(0x1400.0)
print "RxPDO1CommPar.CobID=$%08lX",sdo.r(0x1400.1) ; PDO ID of 1st RPDO    ?
print "RxPDO1CommPar.TrTyp= %03ld",sdo.r(0x1400.2) ; PDO transmission type ?
 ; (tr.type: 255=event  1..240=sync,  253=on RTR    [DS301 10-6] )
print "RxPDO1CommPar.Inhib= %05ld",sdo.r(0x1400.3) ; PDO inhibit time      ?
print "RxPDO1CommPar.CMSpr= %03ld",sdo.r(0x1400.4) ; PDO CMS priority (no) ?
print "RxPDO1CommPar.EvTim= %05ld",sdo.r(0x1400.5) ; PDO event timer [ms]  ?
sdo.w( 0x1400.2, U8, 0x01 ) ; RPDO1 transmission type : "Update outputs on every SYNC"
sdo.w( 0x1400.2, U8, 0x03 ) ; RPDO1 transmission type : "... on every THIRD SYNC" ...
sdo.w( 0x1400.2, U8, 0xFE ) ; RPDO1 transmission type : "async, manufacurer specific"
sdo.w( 0x1400.2, U8, 0xFF ) ; RPDO1 transmission type : "async, as per device profile"
0x080  ; send sync (the "old fashioned one", with ZERO data bytes: 
  ; > The original SYNC message has a length of zero. It contained 0 bytes. 
  ; > A newer SYNC messages which has been introduced with CANopen 4.2 
  ; > includes 1 byte for a SYNC counter.


   ;
   ; 0x1401 = 2nd RPDO Communication parameter
@print "RxPDO2CommPar.Sub0=%ld",sdo.r($1401.0)
@print "RxPDO2CommPar.CobID=$%08lX",sdo.r($1401.1) ; PDO ID of 2nd RPDO    ?
@print "RxPDO2CommPar.TrTyp= %03ld",sdo.r($1401.2) ; PDO transmission type ?
 ; (tr.type: 255=event  1..240=sync,  253=on RTR    [DS301 10-6] )
@print "RxPDO2CommPar.Inhib= %05ld",sdo.r($1401.3) ; PDO inhibit time      ?
@print "RxPDO2CommPar.CMSpr= %03ld",sdo.r($1401.4) ; PDO CMS priority (no) ?
@print "RxPDO2CommPar.EvTim= %05ld",sdo.r($1401.5) ; PDO event timer [ms]  ?
   ;
   ; 0x1600 = 1st RPDO Mapping parameter
@print "RxPDO1Mapping.NrObj= %05ld",sdo.r($1600.0) ; Number of mapped objs ?
@print "RxPDO1Mapping.Obj.1=$%08lX",sdo.r($1600.1) ; First mapped object   ?
@print "RxPDO1Mapping.Obj.2=$%08lX",sdo.r($1600.2) ; 2nd   mapped object   ?
@print "RxPDO1Mapping.Obj.3=$%08lX",sdo.r($1600.3) ; 3rd   mapped object   ?
@print "RxPDO1Mapping.Obj.4=$%08lX",sdo.r($1600.4) ; 4th   mapped object   ?
   ;
   ; 0x1601 = 2nd RPDO Mapping parameter
@print "RxPDO2Mapping.NrObj= %05ld",sdo.r($1601.0) ; Number of mapped objs ?
@print "RxPDO2Mapping.Obj.1=$%08lX",sdo.r($1601.1) ; First mapped object   ?
@print "RxPDO2Mapping.Obj.2=$%08lX",sdo.r($1601.2) ; 2nd   mapped object   ?
@print "RxPDO2Mapping.Obj.3=$%08lX",sdo.r($1601.3) ; 3rd   mapped object   ?
@print "RxPDO2Mapping.Obj.4=$%08lX",sdo.r($1601.4) ; 4th   mapped object   ?
@goto PdoTest  ; switch to "operational" and send a few PDO's ...


;----------------------------------------------------------------------------
; SEHR Spezielle Einstell-Prozeduren .. "DevOnly" !
;----------------------------------------------------------------------------

@stop

@sdo.w($1011.1, U32, 0x64616F6C) ; "load" DEFAULT PARAMETERS (!)
;@sdo.w($1011.1, U32, 0x6C6F6164) ; "load" with wrong signature
000 81 00  ; Node Reset
000 01 00  ; Pre-Operational -> Operational, start PDOs .

print "Device name = %s", sdo.r($1008.0) // device name (a "visible string")


;--------- WB's firmware updater (for IO16-BCCO only) --------------------
; 0x5FE4 = "IBUS Slave Info",  0x5FE5 = "IBUS FW Update", 0x5FE6 = "Master FW Update"
print "Bus Coupler:", sdo.r(0x5FE6)   // what's your firmware, Mr Bus Coupler ?
print "IBUS Slaves:", sdo.r(0x5FE4.0) // how many IBUS-slaves are connected ?
print "Slave1:", sdo.r(0x5FE4.1) // Who are you, and what's your firmware, Slave #1 ?
print "Slave2:", sdo.r(0x5FE4.2) // Who are you, and what's your firmware, Slave #2 ?
print "Slave3:", sdo.r(0x5FE4.3) // Who are you, and what's your firmware, Slave #3 ?
print "Slave4:", sdo.r(0x5FE4.4) // Who are you, and what's your firmware, Slave #4 ?
print "S1_fwu:", sdo.r(0x5FE5.1) // Firmware update command for Slave #1
print "S2_fwu:", sdo.r(0x5FE5.2) // Firmware update command for Slave #2
print "S3_fwu:", sdo.r(0x5FE5.3) // Firmware update command for Slave #3
print "S4_fwu:", sdo.r(0x5FE5.4) // Firmware update command for Slave #4


sdo.w(0x5FE5.1, STR, "Hello Bootloader in Slave #1 !")  // TEST (begins with H : Help)
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "hello Bootloader in the bus coupler !")  // TEST (actually 'h'=help)
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "J")  // non-intuitive NXP ISP command : "J" = Read Part ID (~ device code)
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "K")  // non-intuitive NXP ISP command : "K" = Read Boot Code Version
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "N")  // non-intuitive NXP ISP command : "N" = Read Unique ID
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "R 00000000 40") // WB's FWU-API : Read memory (start,len, max 64 byte)
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "c 00000000 0000003F")   // WB's firmware update API : Calc CRC32 (from,to)
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE6, STR, "e 00000000 0000003F")   // WB's firmware update API : ERASE (from,to)
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command

sdo.w(0x5FE5.01, STR, "x    ")  // WB's firmware update API : "x" = EXIT firmware-update and reboot, Slave #1 !
print "FWU:", sdo.r(0x5FE7,STR) // read the "answer" for the above command


@stop







;----------- IN- and OUTPUT objects... ------------------------------------
 print "NR_DIGIN= %ld * 8bit",sdo.r($6000.0)   ; number of digital inputs
 print "ST_DIGIN= $%02lX",sdo.r($6000.1)       ; state of first 8 digital inputs
 ; Trigger fr das Senden der analogen Inputs per PDO .
 print "DinInterruptEnable= %ld",sdo.r($6005.0) ; read "global interrupt-enable"-flag for d.inputs
 print "DinInterruptMskA = %08b",sdo.r($6006.1) ; trigger mask for 1st 8-bit digital input group
 sdo.w($6005.0,U8,1) ; ENABLE "digital input interrupts" (events for PDO-transmission)
 sdo.w($6005.0,U8,0) ; DISABLE "digital input interrupts"

 ; Digital outputs... 
 print "NR_DIGOUT=%ld * 8bit",sdo.r($6200.0)   ; number of digital outputs
 print "ST_DIGOUT=$%02lX",sdo.r($6200.1)       ; state of first 8 digital outputs
 print "FAULT_MOD=%ld",sdo.r($6206.1) ; read fault mode for 8 outputs
 print "PortDirectionsA = %08b",sdo.r($5FF5.1) ; read port-directions for 1st 8-bit group
 print "PortDirectionsB = %08b",sdo.r($5FF5.2) ; read port-directions for 2nd 8-bit group
 print "DigOutputMaskA  = %08b",sdo.r($6208.1,U8,0) // Filter Mask Digital Output 8-bit
 sdo.w($5FF5.1,U8,0x00) ; DISABLE first 8 digital OUTPUTS
 sdo.w($5FF5.2,U8,0x00) ; DISABLE next 8 digital OUTPUTS
 sdo.w($5FF5.1,U8,0x0F) ; ENABLE only 4 digital OUTPUTS
 sdo.w($5FF5.1,U8,0xFF) ; ENABLE first 8 digital OUTPUTS
 sdo.w($5FF5.2,U8,0xFF) ; ENABLE next 8 digital OUTPUTS (if "bi-directional"...yucc)
 sdo.w($6208.1,U8,0xff) ; allow writing ALL outputs 1..8(4?)  [DS401 "Filter Mask Output 8-Bit" - yucc]
 sdo.w($6200.1,U8,0xff) ; turn outputs 1..8(4?) on via SDO (!)
 sdo.w($6200.1,U8,0x00) ; turn all outputs OFF again
 sdo.w($6200.2,U8,0xff) ; turn outputs 9..16 ON via SDO (!)
 sdo.w($6200.2,U8,0x00) ; turn outputs 9..16 OFF again

 print "Filter Mask 8-bit DO = %8b", sdo.r(0x6208.01, U16)

 ; Test what CiA 401 calls "Change polarity output 8-bit" (0x6202) : 
 sdo.w(0x6202.1, U8, 0xFF ); // invert polarity of digital outputs #1..8
 sdo.w(0x6202.1, U8, 0x00 ); // restore normal polarity of outputs #1..8

 ; Test what CiA 401 calls "Filter mask output 8-bit" (0x6208) :
 ; 1 = output shall be set to the received output value
 ; 0 = output SHALL KEEP THE OLD VALUE (that doesn't mean TURN OFF !)
 sdo.w(0x6208.1, U8, 0x00 ); // freeze old values of digital outputs #1..8
 sdo.w(0x6208.1, U8, 0xFF ); // accept new values for outputs #1..8 (=DEFAULT)


@sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)
000 81 00                        ; Node Reset



 ; Analog inputs & outputs...
 print "NR_AN_IN =%ld",sdo.r($6401.0) ; read number of analog Inputs
 print "AIN1     =%ld",sdo.r($6401.1) ; read 1st 16-bit-analog Input
 print "AIN2     =%ld",sdo.r($6401.2) ; read 2nd 16-bit-analog Input
 print "NR_AN_OUT=%ld",sdo.r($6411.0) ; read number of analog Outputs
 print "AOUT1    =%ld",sdo.r($6411.1) ; read 1st analog Output
 print "AOUT2    =%ld",sdo.r($6411.2) ; read 1st analog Output
 ; Trigger fr das Senden der analogen Inputs per PDO .
 ; ACHTUNG ! REICHLICH KOMPLIZERT - siehe CANopen DS401 Objekt 0x6421 ff + TPDOCommPar !
 ;      
 000 01 00   ; "Start Remode Node" to let PDOs run
 print "AinTrigEnab = %ld", sdo.r($6423.0) ; read "interrupt-enable"-flag for a.inputs
 sdo.w($6423.0,U8,1) ; ENABLE "analog input interrupts" (events for PDO-transmission)
 sdo.w($6423.0,U8,0) ; DISABLE "analog input interrupts"
 print "AinTrigSelN = %ld", sdo.r($6421.0) ; number of analog input trigger selections
 print "AinTrigSel1 = %08b",sdo.r($6421.1) ; analog input trigger selection #1
 print "AinTrigSel2 = %08b",sdo.r($6421.2) ; analog input trigger selection #1
 sdo.w($6421.1,U8,0x01) ; Only select UPPER analog limit to trigger transmission
 print "AinTrigLevLo1 = %ld",sdo.r($6425.1) ; analog input trigger level LO, ch #1
 print "AinTrigLevHi1 = %ld",sdo.r($6424.1) ; analog input trigger level HI, ch #1
 print "AinTrigDelta1 = %ld",sdo.r($6426.1) ; analog input trigger delta, ch #1
 sdo.w($6425.1,I32, +5000)  ;  Set analog input trigger level LOW, ch #1
 sdo.w($6424.1,I32,+29000)  ;  Set analog input trigger level HIGH, ch #1
 sdo.w($6426.1,I32,  2000)  ;  Set analog input trigger "delta", ch #1

@sdo.w($6411.1,I16,0)      ; set 1st analog output to ZERO
@sdo.w($6411.1,I16,16383)  ; set 1st analog output to MID-RANGE
@sdo.w($6411.1,I16,32767)  ; set 1st analog output to MAXIMUM
@sdo.w($6411.2,I16,0)      ; set 2nd analog output to ZERO
@sdo.w($6411.2,I16,16383)  ; set 2nd analog output to MID-RANGE
@sdo.w($6411.2,I16,32767)  ; set 2nd analog output to MAXIMUM


@stop


@sdo.w($1010.1, U32, 0x65766173) ; "save" ALL PARAMETERS (!)


;====== Zustandsuebergaenge des CANopen-Knotens (NMT-Test) ===================
($700+NodeId)rXX ; Anforderung Node-Guard-Status mit RTR (nur mit NEUEM PCAN-USB!)
000 01 00   ; "Start Remode Node" ("Node Start Protocol" in DS203-2 p.7)
            ; CANopen-"RUN"-LED muss auf 'Dauerlicht' schalten.
            ; Die PDOs sollten danach aktiv sein. Test wie folgt:
ctx((P1) 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF) : ctx((P2) 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF 0xFF) ; per PDO alles EIN
ctx((P1) 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00) : ctx((P2) 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00) ; per PDO alles AUS
000 02 00   ; "Stop Remote Node"
000 03 00   ; "Disconnect Remote Node"       (DS203-2 p.8,  kennt SYS TEC wohl nicht)
000 80 00   ; "Enter Preoperational State" -> CANopen-"RUN"-LED *blinkt*. 
000 81 00   ; Operational -> Initialisation (d.h. Reset), danach wieder "Pre-Operational"
#128 0      ; Sync-Telegramm
@stop
@reset  ; reset time, counters and warn-flags of the DOS-CAN-Tester

LaufLicht:  ;====== Ansteuern der 8 digitalen I/Os per PDO =======
 gosub Bootup; "Extended Bootup" wichtig fuer Node-Guarding.
 CommPar := 0x1400;  // find the number of bytes to send in the device's first RPDO..
 gosub GetSizeOfPDO; // in : CommPar,  out : NumBytesMapped[CommPar]
 000 01 00  ; Modul von Pre-Operational -> Operational schalten, PDOs aktiv.
Lauf1:      ; Auch fuer LENZE mit Master/Slave - Kaskadierung (oh Graus...)
 (P1) 01 C0 07 F0
 (P1) 02 60 0E 78
 (P1) 04 30 1C 3C
 (P1) 08 18 38 1E
 (P1) 10 0C 70 0F
 (P1) 20 06 E0 87
 (P1) 40 03 C1 C3
 (P1) 80 01 83 E1
 ($700+NodeId)rXX ; Anforderung Node-Guard-Status via RTR
@goto Lauf1

; hier das ur-alte "Extended NMT Bootup" welches mit SYS TEC's CANopen eh nicht funktioniert:
000 80 00 ; Enter Pre-Operational State
000 03 00 ; Disconnect -> Antwort "#2022"=identify node (kennt SYS TEC nicht, "unknown command")
($700+NodeId)rXX  ; Anforderung Node-Guard-Status (SYS TEC: keine Antwort ? )
(#2026) 04 (N) 00 00 00 00 00 00  ; Connect Remote Node ("Teil 1")
(#2026) 06 00  FF 00 00 00 00 00  ; "Identify Remote Nodes" (!!??!!)
(#2026) 02 (N) (N) 07 00 10 02 02 ; Node Guarding Vorbereitung [203-2p4],"T.2"
(#2026) 03 (N) 01 00 00 00 00 00  ; Prepare Remote Node
@return  ; Ende "ExtBoot"


PdoTest:  ; Test : Switch to "operational" and send something to the RPDO's :
@A=0        ;
000 01 00   ; "Start Remode Node" ("Node Start Protocol" in DS203-2 p.7)
            ; Note: The SYS TEC stack reacts with an EMERGENCY TELEGRAM
            ;       if the count of bytes in an RPDO is not correct ! ! ! 
 $281r XX XX XX XX XX XX XX XX ; request transmission of process data (TPDO2)
 $381r XX XX XX XX XX XX XX XX ; request transmission of process data

PdoTestL:
 $201 FF FF ; -> RPDO1 vom E/A-Modul (digitale Outputs)
 $201 00 00
 $201 11 11
 $201 22 22
 $201 44 44
 $201 88 88
 $301 (A).w (A+0x3FFF).w  ; -> RPDO2 vom E/A-Modul (analoge Outputs)
 @A=(A+1024)&0x7FFF
@goto PdoTestL


; <EOF>
