; SahmLPC1768Test.cts:   Testprogramm zum Ansteuern einer "SAHM-TASTATUR".
;  Lauffhig mit CAN TESTER FR WINDOWS - Start mit  F9  !
; 
;  Ablage: <WoBu5>C:\cbproj\CanTest\Programs\SahmLPC1768Test.TXT
;  Stand:  2010-01-11    Autor: W.Buescher
;
; Eigenschaften bzw. Funktionsweise Tastaturapplikation (laut Email von SAHM): 
; > - default Baudrate 250 kBit/s (wenn Dip-Schalter auf 0 steht!) 
; > - default Node ID 20 dezimal (0x14 hexadezimal) (wenn Dip-Schalter auf 0) 
; > - es ist ein Default PDO Mapping aktiv; d.h. TxPDO 0x194 und RxPDO 0x214
; >    sind konfiguriert
; > - alle 500 ms muss die CAN Message 0x214 mit einer Lnge von 2 Byte gesendet werden 
; > - wird 0x214 nicht alle 500 ms empfangen, leuchten die LEDs in Tasten 1-8 
; >    permanent und die LEDs in Taste 9 und 10 blinken abwechselnd je 3 mal 
; > - 0x214 dient der Ansteuerung der LEDs; ist Bit 0 gesetzt, leuchtet LED 
; >      in Taste 1; Bit 1 -> LED Taste 2 usw.
; > - das Senden von Message 0x194  muss durch einamliges Senden des 
; >      NMT Befehls START_COMMUNICATION  eingeschaltet werden 
; >     (Message 0 mit 2 Bytes; Byte 0 enthlt eine 1 und Byte 1 eine 0x14 (20))
; > - Byte 0 bis 3 der Message 0x194 enthalten den Zustand der Tasten; 
; >      also wenn Taste 1 gedrckt wird, ist in Msg 0x194 Byte 0 eine 1 
; > - Byte 4 bis 7 der Message 0x194 enthalten den Zustand der digitalen 
; >      Eingaenge; also wenn digitaler Eingang 2 ein High hat, 
; >      ist in Byte 4 eine 2 
; > - eine Besonderheit ist der digitale Eingang 5; 
; >      wenn der digitale Eingang 5 eine 0 hat, enthlt Byte 2 der CAN 
; >      Message eine 1 
; > - wenn der digitale Eingang 5 ein High hat, enthlt Byte 2 eine 0 
; >     und Byte 3 eine 0x10 


OnError goto Verflixt ; Sprung falls irgendwas schiefgeht

ClearVars()     ; alte Variableninhalte lschen
; Identifier der verwendeten CAN-Telegramme definieren .
;  Bei SAHM war dies eigentlich immer 0x214 (RPDO) und 0x194(TPDO) .
@D = $14  ; CANopen-hnlicher "Node-ID"
@R = $200+D ; RPDO1 (aus Sicht der Tastatur), SAHM: $214
@T = $180+D ; TPDO1 (aus Sicht der Tastatur), SAHM: $194

@tx_interval=0.01  // set TX interval, also limits execution speed: SECONDS per line

Neustart:
@LedExtraBits = 0;
OpenWindow("SAHM-Tastatur-Test \"LPC 1768\"",w=500,h=333)
SetFont("Arial",20):ClearScreen()
$000 01 14     ; "Start Sendeprotokoll" (SAHM-Slang)
@delay(1)      ; Verzoegerung seit 2006-04 in SEKUNDEN (SI-Einheit)

ClearScreen() 
@LedTestBits = 0 ; keine extra gesetzten LEDs
SetButton(1,"&Neustart",goto Neustart)
SetButton(2,"&Beenden", goto Beenden)
SetButton(3,"&Alle LEDs an", @LedExtraBits=LedExtraBits^0xFFFF)
SetFont("Arial",8):MoveTo(0,10)
print( "CAN-Baudrate = %ld kBit/sec",can.baudrate ) ; 100 oder 250 kBit/sec ?
SetFont("Arial",15,7):MoveTo(10,130)
print( " Nur fr SAHM-Tastatur mit LPC 1768 ! " )
SetFont("Arial",12,0):MoveTo(30,170)
print( "Dieser Test verbindet die Tasten per CAN-Bus mit den LEDs." )
print( "Vor dem Umstecken der CAN-Kabel unbedingt erst den Test " )
print( "beenden (B), und erst nach Anschluss der nchsten Tastatur" )
print( "den Test wieder starten (F9) .. sonst gib't CAN-Trouble !" )
print( "Solange der Test NICHT luft, blinken die \"O\"- und \"I\"-Tasten ." )


$000 01 14       ; "Start Sendeprotokoll" (SAHM-Slang)

Loop1: ; Schleife 1 : fuer BUCHER-Tastatur.
       ; Funktion: Hintergrundhelligkeit aendert sich staendig,
       ; beim Druecken einer Taste muss diese hell aufleuchten.
 @KeyMatrixBits = canrx(T,0,I16,0)  ; 16-bit Tastenzustnde lesen
 @JoystickBits  = canrx(T,16,I16,0) ; 16-bit Joystick-Kontakte
 @DigitalInputs = canrx(T,32,I16,0) ; 16-bit digitale Eingnge
 @LedBits=(LedBits+1)&65535         ; LED-Ansteuer-Bits (2 byte)
 (R) (LedBits | LedExtraBits ).w    ; PDO-Telegramm (0x214) senden
 SetFont("Courier New",26):MoveTo(0,40)
 print "Tast.: %16b",KeyMatrixBits )  ; Matrixbits als Binrzahl
 print "LEDs : %16b",LedBits | LedExtraBits )  ; LED-Test-Muster
@goto Loop1


;--------- Planmaessiges Ende des Tests... ----------------------
Beenden:
CloseWindow()
print "Programmende - Neustart mit F9"
stop

;--------- Bei Fehlern... ---------------------------------------
Verflixt:
 OpenWindow( "CAN-Test : FEHLER ! ",w=400,h=200,bc=red)
 SetFont("Arial",8) : MoveTo(8,4) 
 print( "Fehlercode %ld in Programmzeile %ld", ErrorInfo.code, ErrorInfo.line )
 print( ErrorInfo.text )
 SetFont("Arial",11,2) : MoveTo(8,40) 
 print "- CAN-Baudrate berprfen (momentan %ld kBit/sec)",can.baudrate
 print "- Knotennummer berprfen (momentan #%ld )",sdo.node
 print "- die verdammten Kabel berprfen"
 print "- die DIP-Schalter richtig einstellen"
 SetButton( 1,"&Nochmal versuchen",  goto Neustart )
 SetButton( 2,"&Abbrechen",  goto Abbruch )
Fehlerschleife:
goto Fehlerschleife
Abbruch:  ; ABBRUCH im Fehlerfall, ohne nochmaligen CAN-Sendeversuch !
 CloseWindow()  ; Benutzerdefiniertes Fenster schliessen (falls noch offen)
 print "Programmabbruch - Neustart mit F9 !"
@stop


; EOF